//  ======================================================================== //
//                                                                           //
//    File      : Bmp.h                                                      //
//    Purpose   :                                                            //
//    Time-stamp: <Thu Nov/14/2002 09:34 MET Coder@ReallySoft.de>            //
//                                                                           //
//    (C) February 2002 by Ralf Westram                                      //
//                                                                           //
//    Permission to use, copy, modify, distribute and sell this software     //
//    and its documentation for any purpose is hereby granted without fee,   //
//    provided that the above copyright notice appear in all copies and      //
//    that both that copyright notice and this permission notice appear      //
//    in supporting documentation.                                           //
//                                                                           //
//    Ralf Westram makes no representations about the suitability of this    //
//    software for any purpose.  It is provided "as is" without express or   //
//    implied warranty.                                                      //
//                                                                           //
//  ======================================================================== //

#ifndef BMP_H
#define BMP_H

#ifndef STRING
#include <string>
#endif

#ifndef BOOST_UTIL_H
#include <boost_util.h>
#endif

namespace rs {

    typedef unsigned int   UINT32;
    typedef unsigned short UINT16;

    const long max_distance = 255L*255L;

    /// Reference to a pixel inside a @ref Bmp
    class Pixel {
    private:
        unsigned char *pix;     // pointer to pixel

        static unsigned char transparent_red, transparent_green, transparent_blue; // the transparent color
        static bool          use_transparent_color;

    public:
        /** @brief Construct a NULL-Pixel
        */
        Pixel() : pix(0) {}
        /** @brief Construct a Pixel
            @param pixPtr pointer into Bmp bitmap
            @see Bmp::getPixel
        */
        Pixel(unsigned char *pixPtr) : pix(pixPtr) {}
        virtual ~Pixel() {}

        /// provides low-level acces to the pixel pointer
        unsigned char *getPixPointer() { return pix; }
        /// @overload
        const unsigned char *getPixPointer() const { return pix; }

        /** @brief return one color channel of the pixel
            @param rgb channel selector (0 = red, 1 = green, 2 = blue)
            @return value of the channel
        */
        int color(int rgb) const { assert(rgb >= 0 && rgb <= 2); return pix[rgb]; }

        /** @brief set the color of one channel
            @param rgb channel selector (0 = red, 1 = green, 2 = blue)
            @param value new value for the channel
        */
        void set_color(int rgb, unsigned char value) { assert(rgb >= 0 && rgb <= 2); pix[rgb] = value; }

        /** @brief Tests if a pixel is transparent ( = black)
            @return true if transparent
            @see defineTransparentColor
        */
        bool hasTransparentColor() const {
            return use_transparent_color &&
                color(0) == transparent_red &&
                color(1) == transparent_green &&
                color(2) == transparent_blue;
        }

        /** @brief defines the transparent color (default = 0, 0, 0)
            @param red value for red channel
            @param green value for green channel
            @param blue value for blue channel
        */
        static void defineTransparentColor(unsigned char red, unsigned char green, unsigned char blue);

        /** @brief control whether a transparent color should be used (default is no)
            @param enabled if true = > transparent color will be used
        */
        static void useTransparentColor(bool enabled) {
            use_transparent_color = enabled;
        }

        /** @brief function for fuzzy pixel comparison
            @param other Pixel to compare with
            @param matchCounter if none of the color channels differ more than max_dist matchCounter is incremented
            @param mismatchCounter otherwise mismatchCounter is incremented
            @param max_dist maximum color difference counted as match

            @note if one of the Pixel's @ref hasTransparentColor() the comparison neither counts
            as mismatch nor as match.
        */
        void countMisMatches(const Pixel& other, long& matchCounter, long& mismatchCounter, int max_dist) {
            if (!hasTransparentColor() && !other.hasTransparentColor())  {
                long rdist = labs(color(0)-other.color(0));
                long gdist = labs(color(1)-other.color(1));
                long bdist = labs(color(2)-other.color(2));

                if (rdist>max_dist || gdist>max_dist || bdist>max_dist) {
                    ++mismatchCounter;
                }
                else {
                    ++matchCounter;
                }
            }
        }

        /** @brief calculates the color-distance between two Pixels
            @param other the other pixel
            @return distance color distance
        */
        long distance(const Pixel& other) const {
            if (hasTransparentColor() || other.hasTransparentColor()) return 0;

            long rdist = color(0)-other.color(0);
            long gdist = color(1)-other.color(1);
            long bdist = color(2)-other.color(2);

            return rdist*rdist + gdist*gdist + bdist*bdist;
        }

        /** @brief increments to the next pixel (rightwards, line by line)
            @note be careful: you may run out of the bitmap with this operator
         */
        Pixel& operator++() {
            pix += 3;
            return *this;
        }

        bool operator == (const Pixel& other) const {
            return
                color(0) == other.color(0) &&
                color(1) == other.color(1) &&
                color(2) == other.color(2);
        }

        bool operator != (const Pixel& other) const { return !(*this == other); }
    };




    /// a windows bitmap (might be loaded and saved)
    class Bmp {
    private:
        std::string loaded_from_file;

        UINT32 xsize;
        UINT32 ysize;

        unsigned char *data;
        size_t         offset_per_line;

        static void throw_wrong_pix_pos();

        unsigned char *pixelPointer(UINT32 xpos, UINT32 ypos) {
            if (xpos >= xsize || ypos >= ysize) throw_wrong_pix_pos();
            return data + ypos*offset_per_line + xpos*3;
        }

        void getPositions(const Pixel& pixel, UINT32& xpos_result, UINT32& ypos_result) const {
            long distance = pixel.getPixPointer()-data;
            long ypos     = distance/long(offset_per_line);
            long xpos     = (distance - ypos*offset_per_line)/3;

            xpos_result = xpos;
            ypos_result = ypos;

            if (xpos_result >= xsize || ypos_result >= ysize) throw_wrong_pix_pos();
        }

    public:
        /** @brief read a bitmap from %file
            @param filename guess what
        */
        Bmp(const std::string& filename);
        /** @brief construct an empty bitmap
            @param xsize_ width of bitmap
            @param ysize_ height of bitmap
            @note All pixels are set to color (0, 0, 0)
        */
        Bmp(size_t xsize_, size_t ysize_);

        Bmp(const Bmp& other);
        Bmp& operator=(const Bmp& other);

        virtual ~Bmp() { delete data; }

        /// get width of bitmap
        UINT32 Xsize() const { return xsize; }

        /// get height of bitmap
        UINT32 Ysize() const { return ysize; }

        /** @brief get access to one pixel of the bitmap
            @param xpos horizontal offset in the bitmap (from left to right)
            @param ypos vertical offset in the bitmap (from top to bottom)
            @return the Pixel
        */
        Pixel getPixel(UINT32 xpos, UINT32 ypos)  { return Pixel(pixelPointer(xpos, ypos)); }

        /// @overload
        const Pixel getPixel(UINT32 xpos, UINT32 ypos) const { return Pixel(const_cast<Bmp*>(this)->pixelPointer(xpos, ypos)); }

        /** @brief set the value of a pixel in the Bmp
            @param xpos horizontal position
            @param ypos vertical position
            @param red value for red channel
            @param green value for green channel
            @param blue value for blue channel
        */
        void setPixel(UINT32 xpos, UINT32 ypos, unsigned char red, unsigned char green, unsigned char blue) {
            unsigned char *d = const_cast<unsigned char *>(pixelPointer(xpos, ypos));

            d[0] = red;
            d[1] = green;
            d[2] = blue;
        }

        /** @brief save the bitmap to %file
            @param filename name of %file to save
        */
        void saveAs(const std::string& filename) const;

        /** @brief returns the filename the Bmp was loaded from
            @returns filename (or  "" if Bmp wasn't loaded)
         */
        const std::string& originalFilename() const { return loaded_from_file; }

        /// move a Pixel leftwards
        void left(Pixel& curr, size_t dist = 1) {
            UINT32 xpos, ypos;
            getPositions(curr, xpos, ypos);
            curr = getPixel(xpos-dist, ypos);
        }

        /// move a Pixel rightwards
        void right(Pixel& curr, size_t dist = 1) {
            UINT32 xpos, ypos;
            getPositions(curr, xpos, ypos);
            curr = getPixel(xpos+dist, ypos);
        }

        /// move a Pixel upwards
        void up(Pixel& curr, size_t dist = 1) {
            UINT32 xpos, ypos;
            getPositions(curr, xpos, ypos);
            curr = getPixel(xpos, ypos-dist);
        }

        /// move a Pixel downwards
        void down(Pixel& curr, size_t dist = 1) {
            UINT32 xpos, ypos;
            getPositions(curr, xpos, ypos);
            curr = getPixel(xpos, ypos+dist);
        }
    };

    // create a new BMP with half width and height
    Bmp *halfBmp(const Bmp& bmp);



}; // end of namespace rs

#else
#error Bmp.h included twice
#endif // BMP_H

