//  ======================================================================== //
//                                                                           //
//    File      : Date.h                                                     //
//    Purpose   :                                                            //
//    Time-stamp: <Wed Dec/26/2001 14:02 MET Coder@ReallySoft.de>            //
//                                                                           //
//    (C) December 2001 by Ralf Westram                                      //
//                                                                           //
//    Permission to use, copy, modify, distribute and sell this software     //
//    and its documentation for any purpose is hereby granted without fee,   //
//    provided that the above copyright notice appear in all copies and      //
//    that both that copyright notice and this permission notice appear      //
//    in supporting documentation.                                           //
//                                                                           //
//    Ralf Westram makes no representations about the suitability of this    //
//    software for any purpose.  It is provided "as is" without express or   //
//    implied warranty.                                                      //
//                                                                           //
//  ======================================================================== //

#ifndef DATE_H
#define DATE_H

#ifndef CTIME
#include <ctime>
#endif

namespace rs {

    /** @memo Date related functions with calculation of german holidays
     */
    namespace date {

        // --------------------
        //      class Date
        // --------------------
        /// Date stores date and time in a 'struct tm'
        class Date {
        private:
            std::time_t time;
            std::tm     date;
            bool        valid;

            void init(const std::time_t& time_);
            void init(const std::tm& date_);
            void init(int day, int month, int year);

            void correct();

        public:
            /** create the current Date */
            Date() { std::time(&time); init(time); }

            /** create a new date (clock is set to 00:00.00 ) */
            Date(int day_, int month_, int year_) { init(day_, month_, year_); }

            /** create a new date from struct tm */
            Date(const std::tm& date_) { init(date_); }

            /** create a new date from another date */
            Date(const Date& other) {
                time  = other.time;
                date  = other.date;
                valid = other.valid;
            }

            /** assign a Date */
            Date& operator = (const Date& other) {
                if (&other != this) {
                    time  = other.time;
                    date  = other.date;
                    valid = other.valid;
                }
                return *this;
            }

            virtual ~Date() {}

            /** returns true if the Date is a valid Date */
            bool is_valid() const { return valid; }

            /** returns date as 'struct tm' */
            const std::tm& get_Date() const { return date; }

            /** returns zero if date is no 'Feiertag' (= german holiday), index otherwise */
            int is_Feiertag() const;

            /** increment date
                @param offset days to increment
             */
            void inc(int offset = 1) { date.tm_mday += offset; correct(); }
            /** decrement date
                @param offset days to decrement
             */
            void dec(int offset = 1) { date.tm_mday -= offset; correct(); }

            /** returns year */
            int year() const { return get_Date().tm_year+1900; }
            /** returns month */
            int month() const { return get_Date().tm_mon+1; }
            /** returns day of month */
            int day() const { return get_Date().tm_mday; }

            /** returns day of week (0 for sunday) */
            int dayOfWeek() const { return get_Date().tm_wday; }
        };

        // some useful date functions :

        /** returns true if year is a leap year
            @param year the year to test
         */
        inline bool is_leap_year(int year) { return Date(29, 2, year).is_valid(); }

        /** returns the name of a 'Feiertag'
            @param feiertag_idx allowed values: 1..n (should be result from is_Feiertag())
         */
        const char *name_of_Feiertag(int feiertag_idx);

        /** returns true if Feiertag is catholic
            @param feiertag_idx allowed values: 1..n (should be result from is_Feiertag())
         */
        bool is_catholic(int feiertag_idx);

    }; // end of namespace date
}; // end of namespace rs

#else
#error Date.h included twice
#endif // DATE_H

