/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000
// Ralf Westram
// Time-stamp: <Fri Dec/06/2002 05:19 MET Coder@ReallySoft.de>
//
// Permission to use, copy, modify, distribute and sell this software
// and its documentation for any purpose is hereby granted without fee,
// provided that the above copyright notice appear in all copies and
// that both that copyright notice and this permission notice appear
// in supporting documentation.  Ralf Westram makes no
// representations about the suitability of this software for any
// purpose.  It is provided "as is" without express or implied warranty.
//
// This code is part of my library.
// You may find a more recent version at http://www.reallysoft.de/
//
/////////////////////////////////////////////////////////////////////////////

#ifndef DIRECTORY_H
#define DIRECTORY_H

#ifndef __SET__
#include <set>
#endif
#ifndef __MAP__
#include <map>
#endif

// #define HAVE_STDC
#ifndef __STAT_H__
#include <sys/stat.h>
#endif

#ifndef BINFILE_H
#include <BinFile.h>
#endif


namespace rs {

    /** @brief Files and Directories

        @see rs::posix for low-level %file system operations
     */
    namespace file {

        /// contains the separator used in full pathnames
        extern char DIR_SEPARATOR;

        /** @brief concatenate directory with filename
            @param directory name of directory
            @param filename name of %file
            @return concatenated name */
        std::string catDF(const std::string& directory, const std::string& filename);

        //  -------------------------------------------------
        //      class Directory : public BinaryStoreable
        //  -------------------------------------------------

        /** @brief class containing all or some files of a directory

            @warning You should not use this class directly.
            Use @ref Directories and @ref scanDir instead - even if you don't want to recurse
            a whole directory tree.
        */
        class Directory : public rs::file::BinaryStoreable {
        private:
            std::string name_; // full path of Directory
            std::set<std::string> files; // the files in this directory
            time_t oldest_; // modification time of oldest
            time_t newest_; // and newest accepted file

        public:
            /// Create a Directory
            Directory() { oldest_ = newest_ = 0; }
            virtual ~Directory() {}

            /** @brief adds a filename to Directory
                @param filename the filename
                @param t time associated with %file (normally time of last modification) */
            void addFile(const std::string& filename, time_t t);

            /** @brief change directory name
                @param dirname new name of directory */
            void setName(const std::string& dirname) { name_ = dirname; }

            /// get name of directory
            const std::string& name() const { return name_; }

            /// get number of files in directory
            size_t size() const { return files.size(); }

            /// iterator for files in directory
            std::set<std::string>::const_iterator begin() const { return files.begin(); }
            /// iterator for files in directory
            std::set<std::string>::const_iterator end() const { return files.end(); }

            /** @brief return full path name of a %file
                @param filename of the %file
                @return full path of %file with directory name */
            std::string fullPath(const std::string& filename) const;

            /// get time of oldest %file in directory
            time_t oldest() const { return oldest_; }
            /// get time of newest %file in directory
            time_t newest() const { return newest_; }

            /// save directory into BinaryFile
            virtual void save(rs::file::BinaryFile& out) const;
            /// load directory from BinaryFile
            virtual void load(rs::file::BinaryFile& in, int version);
        };

        /** @brief Contains several Directory's

            Normally Directories should be initialized by using scanDir()
         */
        typedef std::map<std::string, Directory> Directories;

        /// Predicate for files/directories
        typedef bool (*predicate)(const std::string& dirname, const char *filename, const struct stat& filestat);

        /** @brief scan content of directories (recursion controllable)

            Here you can influence the way how recursion is done.

            @param dirs scanned directories are added here
            @param rootdir path of root directory
            @param recurse if true then recurse subdirs
            @param follow_links if true then follow links (not implemented yet!)
            @param takeFile callback returning true if %file should be added
            @param recurseDirectory callback returning true if directory should be scanned recursive
            @param display if non-NULL: stream where progress info should be displayed
        */
        void scanDir(Directories&        dirs,
                     const std::string&  rootdir,
                     bool                recurse,
                     bool                follow_links,
                     predicate           takeFile,
                     predicate           recurseDirectory,
                     std::ostream       *display);

        /** @brief scan content of directories

            @overload
        */
        void scanDir(Directories&        dirs,
                     const std::string&  rootdir,
                     bool                recurse,
                     bool                follow_links,
                     predicate           takeFile,
                     std::ostream       *display);

        /// save directory container into a %file
        void saveDirectories(const Directories& dirs, rs::file::BinaryFile& out);
        /// load directory container from a %file
        void loadDirectories(Directories& dirs, rs::file::BinaryFile& in);

        /** @brief create canonical filename
            @param filename relative filename
            @param to_win32 if true the filename is converted to WINDOWS path format (defaults to true if DIR_SEPARATOR == \)
            @return the absolute (canonical) version of filename
         */
        std::string canonicalize(const std::string& filename, bool to_win32 = (DIR_SEPARATOR == '\\'));

        /** @brief test if filename is canonical
            @return true if @p filename is canonical
         */
        inline bool is_canonical(const std::string& filename) { return filename == canonicalize(filename); }

        /** @brief extract directory part from filename
            @param fullpath full pathname
            @return the directory part of @p fullpath
        */
        std::string extractDirectory(const std::string& fullpath);

        /** @brief remove directory part from a filename
            @param fullpath full pathname
            @return the filename of @ fullpath w/o directory information
        */
        std::string extractFilename(const std::string& fullpath);

        /// converts all backslashes to slashes
        void backslashes2slashes(std::string& filename);

        /** @brief test if a %file is a windows link (*.lnk)
            @param file filename of the potential link
            @param target the path of the link target is saved here (only if @p true is returned)
            @return true if %file is a link
        */
        bool isLink(const std::string& file, std::string& target);
    };

}; // end of namespace rs

#else
#error Directory.h included twice
#endif // DIRECTORY_H
