/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000
// Ralf Westram
// Time-stamp: <Sun Nov/24/2002 18:17 MET Coder@ReallySoft.de>
//
// Permission to use, copy, modify, distribute and sell this software
// and its documentation for any purpose is hereby granted without fee,
// provided that the above copyright notice appear in all copies and
// that both that copyright notice and this permission notice appear
// in supporting documentation.  Ralf Westram makes no
// representations about the suitability of this software for any
// purpose.  It is provided "as is" without express or implied warranty.
//
// This code is part of my library.
// You may find a more recent version at http://www.reallysoft.de/
//
/////////////////////////////////////////////////////////////////////////////

#ifndef ERR_H
#define ERR_H

#ifndef RS_ASSERT_H
#include <rs_assert.h>
#endif
#ifndef BOOST_UTIL_H
#include <boost_util.h>
#endif



namespace rs {

    /** @brief Error-classes used for exceptions.

        Derive your own exception classes from Error and overload get_print_message
        to customize the error message.

        You should edit MAIL_CONTACT to fit to your needs.
     */
    namespace err {

        /// Base exception class for programm errors.
        class Error : public std::runtime_error {
        protected:
            std::string        error_module;
            static std::string module_name;
            bool               real_error;
        public:
            /** Create a new error.
                @param s Error message
                @param real_error_ true -> 'Error in ..'
                @see rs::err
            */
            Error(bool real_error_ /*= true*/, const std::string& s);

            /** Create a new error.
                @overload
                @note like above with real_error == true */
            Error(const std::string& s);

            /** Create a new error.
                @param s Error message
                @param real_error_ true -> 'Error in ..' */
            Error(bool real_error_ /*= true*/, const char *s);

            /** Create a new error.
                @overload
                @note like above with real_error == true */
            Error(const char *s);

            virtual ~Error() throw();

            /** Print error message.
                @param out Stream to which the error is printed
                @see get_print_message */
            virtual void print(std::ostream& out) const;

            /** Get error message as it would be printed by print.
                @return String containing the error message: "Error in <module>: <message>"
                @see print
                @see set_module_name
            */
            virtual std::string get_print_message() const;

            /** Get plain error message
                @return const char* containing the plain error message. */
            virtual const char *what() const throw() { return runtime_error::what(); }

            /** Get plain error message.
                @return string containing the plain error message. */
            const char * get_message() const { return what(); }

            /** Changes the module name shown in error messages.
                @param new_module_name Name of current module
                @return Name of previous module */
            static std::string set_module_name(const std::string& new_module_name);

            friend void reThrow(const exception& e);
        };

        /* use this to rethrow an exception that may be an Error.

           Error and derived classes are copied and thrown.
           Other exceptions are thrown as InternalError */
        void reThrow(const std::exception& e);

        // ---------------------------------------------------
        //      class ErrorModule : rs_boost::noncopyable
        // ---------------------------------------------------
        /// An easy way to define a new error-source
        class ErrorModule : rs_boost::noncopyable {
        private:
            std::string old_module_name;

        public:
            /** Changes the module name used by Error.

                Define one instance of ErrorModule in the main block of your module.

                If you are using ProgramInfo then you should be aware, that it uses set_module_name too.
                It is intended that you call ProgramInfo::init before you call ErrorModule.

                @param  new_module_name     name of module
                @param  appendTo            if true then append to existing module_name (default)

                @see        Error::set_module_name
                @see        ProgramInfo::init
            */
            explicit ErrorModule(const std::string& new_module_name, bool appendTo=true) {
                old_module_name = Error::set_module_name(new_module_name);
                if (appendTo && !old_module_name.empty()) {
                    Error::set_module_name(old_module_name+"::"+new_module_name);
                }
            }
            ~ErrorModule() {
                Error::set_module_name(old_module_name);
            }
        };


        // --------------------------------------------------------------------------------
        //     class InternalError : public Error
        // --------------------------------------------------------------------------------
        /// Exception class for internal (fatal) program errors.
        class InternalError : public Error {
        public:
            /** Create a new internal error.
                @param s Error message */
            InternalError(const std::string& s) : Error(s) {
#if !defined(FINAL_VERSION)
                cerr << "Throwing InternalError(" << s << ")\n";
#endif // FINAL_VERSION
            }
            virtual ~InternalError() throw() {}

            /** Get error message.
                @return String containing the error message:
                "InternalError in <module>: <message> (Please report ...)"
                @see MAIL_CONTACT
                @see set_module_name
                @see print */
            virtual std::string get_print_message() const;
        };

        // --------------------------------------------------------------------------------
        //     class IOError : public Error
        // --------------------------------------------------------------------------------
        /// Exception class for I/O errors.
        class IOError : public Error {
        public:
            /** Create a new I/O error.

                Example: IOError("my.txt") gives sth like "permission denied ('my.txt')"

                @param filename Name of the %file for which an I/O operation failed.
            */
            IOError(const std::string& filename);

            /** Create a new I/O error.

                Example: IOError("my.txt", "unlink") gives sth like "permission denied for unlink('my.txt')"

                @param filename Name of the %file for which an I/O operation failed.
                @param IO_command Name of the failed IO_command
            */
            IOError(const std::string& filename, const std::string& IO_command);

            virtual ~IOError() throw() {}
        };

    };
};

#else
#error Err.h included twice
#endif // ERR_H
