/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000
// Ralf Westram
// (Coded@ReallySoft.de)
//
// Permission to use, copy, modify, distribute and sell this software
// and its documentation for any purpose is hereby granted without fee,
// provided that the above copyright notice appear in all copies and
// that both that copyright notice and this permission notice appear
// in supporting documentation.  Ralf Westram makes no
// representations about the suitability of this software for any
// purpose.  It is provided "as is" without express or implied warranty.
//
// This code is part of my library.
// You may find a more recent version at http://www.reallysoft.de/
//
/////////////////////////////////////////////////////////////////////////////

#ifndef INPUTFILE_H
#define INPUTFILE_H

#ifndef __CSTDIO__
#include <cstdio>
#endif

#ifndef ERR_H
#include "Err.h"
#endif
#ifndef TOOLS_H
#include "Tools.h"
#endif

namespace rs {
    namespace file {

#define RS_INPUTFILE_BUFSIZE 2048 // max size of a line

        // --------------------------------------------------------------------------------
        //     class InputFile
        // --------------------------------------------------------------------------------
        /// read textfiles line per line and provide error raising
        class InputFile {
        private:
            std::string    filename;
            unsigned long  current_line;
            FILE          *in;
            size_t         bufsize;
            size_t         offset; // read offset into buffer
            size_t         filled; // no of bytes in buffer
            char          *buffer;
            bool           end_of_file_seen;
            bool           ignore_empty;
            int            do_auto_break; // 0 -> no autobreak
            std::string    breakAt; // inside this string autobreak is allowed
            int            breakAtPos; // at this position of 'breakAt'

            void fillBuffer() {
                assert(!end_of_file_seen);
                filled = fread(buffer, 1, bufsize, in);
                if (filled<bufsize) end_of_file_seen = true;
                offset = 0;
            }

        public:
            /** @brief open a %file for reading
                @param filename_    name of %file to read
                @param bufsize_     size of line buffer has to be longer than longest line [#Default: 2048#]
                @exception IOError
            */
            InputFile(const std::string& filename_, int bufsize_=RS_INPUTFILE_BUFSIZE);

            /** @brief open STDIN for reading
                @param bufsize_     size of line buffer has to be longer than longest line [#Default: 2048#]
            */
            InputFile(int bufsize_=RS_INPUTFILE_BUFSIZE);

            /// close the InputFile
            virtual ~InputFile() {
                delete buffer;
                if (in) fclose(in);
            }

            /** @brief defines a substring where lines may be broken automatically.
                @param max_line_length      wanted length of lines (may be ignored by getLine)
                @param toBreak              a string where line may be broken (i.e. "><" for HTML-Files)
                @param toBreakAtPos         the offset inside 'toBreak' where linebreak will happen (i.e. 1 for example above)

                @code
                InputFile html("index.html");
                html.autoLinebreak(80, "><", 1);

                InputFile text("example.txt");
                text.autoLinebreak(80, " ", 1);
                @endcode

                @todo use two parameters for @p toBreakAtPos to allow to eliminate the
                space which is leftover at the end of broken line (in the last example)
            */
            void autoLinebreak(int max_line_length, const std::string& toBreak, int toBreakAtPos);

            /** @brief read next line from %file
                @param result after success this contains the content of the line
                @return #true# if successful
            */
            bool getLine(std::string& result);

            /** @brief get number of last line read by getLine
                @return number of line [1..n]
            */
            unsigned long getLineNumber() const { return current_line; }

            /** @brief what to do with empty lines
                @param ignore if true empty lines are overread
            */
            void ignoreEmptyLines(bool ignore) { ignore_empty = ignore; }

            /** @brief raise an Error referencing the current linenumber.

                The raised Error contains sth like
                'Error in filename (##linenumber): message'

                @param message error message to raise
                @exception err::Error (always thrown)
            */
            void throw_Error(const std::string& message) const {
                throw err::Error(str::strf("Error in %s (#%i): %s", filename.c_str(), current_line, message.c_str()));
            }
        };

    };
};

#else
#error InputFile.h included twice
#endif // INPUTFILE_H
