/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000
// Ralf Westram
// (Coded@ReallySoft.de)
//
// Permission to use, copy, modify, distribute and sell this software
// and its documentation for any purpose is hereby granted without fee,
// provided that the above copyright notice appear in all copies and
// that both that copyright notice and this permission notice appear
// in supporting documentation.  Ralf Westram makes no
// representations about the suitability of this software for any
// purpose.  It is provided "as is" without express or implied warranty.
//
// This code is part of my library.
// You may find a more recent version at http://www.reallysoft.de/
//
/////////////////////////////////////////////////////////////////////////////

#include "Option.h"
#include "Tools.h"
#include "Directory.h"
#include "sys_dep.h"
#include "path.h"

#include <iomanip>
#include <algorithm>
#include <functional>
#include <cstring>

#if defined(DUMP)
//#define DUMP_PARSE
#endif // DUMP

#if defined(DEBUG)
// #define PATH_DETECT
#endif // DEBUG

using namespace std;
using namespace rs;
using namespace rs::cmdline;
using namespace rs::err;
using namespace rs::str;
using namespace rs::file;
using namespace rs::posix;

typedef list<CmdLineOption*> Options;
typedef list<CmdLineArgument*> Arguments;
typedef list<CmdLineAnnotation*> Annotations;

static Options *options = 0;
static Arguments *arguments = 0;
static Annotations *annotations = 0;

static bool initialized = false; // was ProgramInfo::init called ?
static size_t longest_arg_name_length = 6; // 6=length("<opts>")
static size_t longest_optarg_name_length = 0;

static bool options_are_case_sensitive = false; // will be automatically set to true, if options
                                                // with same letter but different case occur

static Terminator *there_can_only_be_one = 0; // will install my own terminator function (real_terminate)

// --------------------------------------------------------------------------------
//     inline void test_initialized(string where_tested)
// --------------------------------------------------------------------------------
inline void test_initialized(string where_tested) {
    if (!initialized) {
        string msg = "ProgramInfo::init() not called in "+where_tested;
        throw InternalError(msg);
    }
}

// --------------------------------------------------------------------------------
//     class InfoError : public Error
// --------------------------------------------------------------------------------

/// raised Errors concerned with command line usage
class InfoError : public Error {
    const ProgramInfo& info;
public:
    InfoError(const ProgramInfo& info_, bool real_error_ /*= true*/, const string& msg_)
        : Error(real_error_, msg_),
          info(info_)
    {
    }
    InfoError(const ProgramInfo& info_, const string& msg_)
        : Error(true, msg_),
          info(info_)
    {
    }
    virtual ~InfoError() throw() {}

    virtual void print(ostream& out) const {
        info.infos(out);
        Error::print(out);
    }
};

//  -----------------------------------------
//      class StringError : public Error
//  -----------------------------------------

/// an Error with an extra message attached
class StringError : public Error {
    string s;
public:
    StringError(const string& msg_, const string& extra_message)
        : Error(true, msg_)
        , s(extra_message)
    {
    }
    virtual ~StringError() throw() {}

    virtual void print(ostream& out) const {
        Error::print(out);
        out << s;
    }
};

// --------------------------------------------------------------------------------
//     CmdLineOption::CmdLineOption(char optChar_, const char *helptext_)
// --------------------------------------------------------------------------------
CmdLineOption::CmdLineOption(char optChar_, const char *helptext_)
    : CmdLineHelpText(helptext_)
{
#if defined(DUMP_PARSE)
    cout << "CmdLineOption(" << optChar_ << ", " << helptext_ << ")\n";
#endif // DUMP_PARSE
    used = false;
    optChar = optChar_;
    if (!options) options = new Options();
    options->push_back(this);
}

CmdLineOption::~CmdLineOption() {
}

void CmdLineOption::print(ostream& out) const {
    out << '-' << optChar << string(longest_optarg_name_length+1, ' ');
}

//  ---------------------------------------------------------------------------------
//      CmdLineAnnotation::CmdLineAnnotation(const char *helptext_, bool spaced_)
//  ---------------------------------------------------------------------------------
CmdLineAnnotation::CmdLineAnnotation(const char *helptext_, bool spaced_)
    : CmdLineHelpText(helptext_)
{
    print_infrontof_option = options ? options->size() : 0;
    if (!annotations) annotations = new Annotations();
    annotations->push_back(this);
    spaced = spaced_;
}

// --------------------------------------------------------------------------------
//     CmdLineArgumentOption::CmdLineArgumentOption(char optChar_, const char *name_, const char *helptext_, const char *value_)
// --------------------------------------------------------------------------------
CmdLineArgumentOption::CmdLineArgumentOption(char optChar_, const char *name_, const char *helptext_, const char *value_)
    : CmdLineOption(optChar_, helptext_)
    , CmdLineArgument(name_, helptext_, value_, false)
    , CmdLineHelpText(helptext_)
{
#if defined(DUMP_PARSE)
    cout << "CmdLineArgumentOption(" << optChar_ << ", " << name_ << ", " << helptext_ << ", " << value_ << ")\n";
#endif // DUMP_PARSE
    if (value_) {
        if (value_[0]==0) {
            cerr << "Empty default argument for option -" << optChar_ << endl;
            throw InternalError("see above");
        }
        CmdLineOption::set();
    }
    longest_optarg_name_length = max(get_name().length(), longest_optarg_name_length);
}

// --------------------------------------------------------------------------------
//     CmdLineArgumentOption::~CmdLineArgumentOption()
// --------------------------------------------------------------------------------
CmdLineArgumentOption::~CmdLineArgumentOption() {
}

// --------------------------------------------------------------------------------
//     void CmdLineArgumentOption::print(ostream& out) const
// --------------------------------------------------------------------------------
void CmdLineArgumentOption::print(ostream& out) const {
    out << '-' << get_optChar() << get_name() << string(longest_optarg_name_length-get_name().length()+1, ' ');
}

//  ------------------------------------------
//      int CmdLineArgument::getNumeric()
//  ------------------------------------------
int CmdLineArgument::getNumeric() {
    string val    = get();
    int    digits = 0;
    int    xs     = 0;
    int    len    = val.length();

    int p;
    for (p=0; p<len; ++p) {
        char c = val[p];
        if (c>='0' && c<='9') {
            ++digits;
        }
        else {
            if (c=='x') ++xs;
        }
    }

    int result = 0;
    if (len==digits) { // only digits found
        result = atoi(val.c_str());
    }
    else {
        if ((xs+digits)==len && xs==1 && val[0]=='0' && val[p]=='x') { // hex-number
            result = strtoul(val.c_str()+2, 0, 16);
        }
        else {
            throw Error(str::strf("Numeric argument awaited in <%s> '%s' (dec or hex '0x..') ",
                                  name.c_str(), val.c_str()));
        }
    }

    return result;
}

//  --------------------------------------------
//      double CmdLineArgument::getDouble()
//  --------------------------------------------
double CmdLineArgument::getDouble() {
    string val    = get();
    int    digits = 0;
    int    kommas = 0;
    int    len    = val.length();

    int p;
    for (p=0; p<len; ++p) {
        char c = val[p];
        if (c>='0' && c<='9') {
            ++digits;
        }
        else {
            if (c=='.') ++kommas;
            else if (c==',') { ++kommas; val[p] = '.'; }
        }
    }

    double result = 0;
    if (len==digits) { // only digits found
        result = atoi(val.c_str());
    }
    else {
        if ((kommas+digits) == len && kommas==1) { // floating point
            result = atof(val.c_str());
        }
        else {
            throw Error(str::strf("Floating point argument awaited in <%s> '%s'",
                                  name.c_str(), val.c_str()));
        }
    }

    return result;
}

// --------------------------------------------------------------------------------
//     CmdLineArgument::CmdLineArgument(const char *name_, const char *helptext_, const char *value_, bool is_real_arg)
// --------------------------------------------------------------------------------
CmdLineArgument::CmdLineArgument(const char *name_, const char *helptext_, const char *value_, bool is_real_arg)
    : CmdLineHelpText(helptext_),
      name(name_),
      value(value_ ? value_ : ""),
      has_default_value(false),
      really_used(false)
{
#if defined(DUMP_PARSE)
    cout << "CmdLineArgument(" << name_ << ", " << helptext_ << ", " << int(is_real_arg) << ")\n";
#endif // DUMP_PARSE
    if (is_real_arg) {
        if (!arguments) arguments = new Arguments(0);
        arguments->push_back(this);
        longest_arg_name_length = max(get_name().length(), longest_arg_name_length);
    }
}

namespace rs {
    namespace cmdline {
        //  ------------------------------------------------------
        //      bool acceptAnyArgument(const string */*arg*/)
        //  ------------------------------------------------------
        bool acceptAnyArgument(const string */*arg*/) {
            return true;
        }
    };
};

// --------------------------------------------------------------------------------
//     CmdLineArgument::~CmdLineArgument()
// --------------------------------------------------------------------------------
CmdLineArgument::~CmdLineArgument() {
}

// start of implementation of class RepeatableCmdLineArgument:

// ------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
//      RepeatableCmdLineArgument::RepeatableCmdLineArgument(const char *name_, const char *helptext_, int min_occur_, int max_occur_, CmdLineArgumentPredicate isArg_, bool is_real_arg)
// ------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
RepeatableCmdLineArgument::RepeatableCmdLineArgument(const char *name_, const char *helptext_, int min_occur_, int max_occur_, CmdLineArgumentPredicate isArg_, bool is_real_arg)
    : CmdLineArgument(name_, helptext_, 0, is_real_arg)
    , CmdLineHelpText(helptext_)
{
    min_occur = min_occur_;
    max_occur = max_occur_;
    isArg = isArg_;
    count = 0;

    // do not set longest_arg_name_length here (it's set in cons of CmdLineArgument)
    // and we don't want to know the length of '[<arg>]+' but of '<arg>'

    //longest_arg_name_length = max(get_name().length(), longest_arg_name_length);
}

// -----------------------------------------------------------
//      string RepeatableCmdLineArgument::get_name() const
// -----------------------------------------------------------
string RepeatableCmdLineArgument::get_name() const {
    string Name = string("[") + CmdLineArgument::get_name() + "]";

    if (min_occur==0) {
        if (max_occur>1) Name = Name+"*";
    }
    else {
        Name = Name+"+";
    }
    return Name;
}

// ------------------------------------------------
//      string RepeatableCmdLineArgument::get()
// ------------------------------------------------
string RepeatableCmdLineArgument::get()
{
    string result = CmdLineArgument::get();
    if (!result.empty()) seen_values.push_back(result); // store as seen
    if (repeated_values.empty()) {
        CmdLineArgument::set("");
    }
    else {
        CmdLineArgument::set(repeated_values.front());
        repeated_values.pop_front();
    }
    return result;
}

// ------------------------------------------------
//      void RepeatableCmdLineArgument::reset()
// ------------------------------------------------
void RepeatableCmdLineArgument::reset()
{
    while (!get().empty()) ;    // read leftover arguments
    repeated_values = seen_values;
    seen_values.clear();
    get();
}

// -end- of implementation of class RepeatableCmdLineArgument.

// --------------------------------------------------------------------------------
//     class HasOptChar : public unary_function<CmdLineOption*, char>
// --------------------------------------------------------------------------------
/// Predicate to search for the corresponding CmdLineOption
class HasOptChar : public unary_function<CmdLineOption*, char> {
    char optChar;
public:
    explicit HasOptChar(char optChar_) :optChar(optChar_) {}
    bool operator() (const CmdLineOption* const & opt) const { return optChar==opt->get_optChar(); }
};

// start of implementation of class RepeatedCmdLineArgumentOption:

// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------
//      RepeatedCmdLineArgumentOption::RepeatedCmdLineArgumentOption(char optChar_, const char *argname_, char separator_, const char *helptext_, const char *value_)
// ----------------------------------------------------------------------------------------------------------------------------------------------------------------------
RepeatedCmdLineArgumentOption::RepeatedCmdLineArgumentOption(char optChar_, const char *argname_, char separator_, const char *helptext_, const char *value_)
    : CmdLineOption(optChar_, helptext_)
    , RepeatableCmdLineArgument(argname_, helptext_, 1, -1, acceptAnyArgument, false)
    , CmdLineHelpText(helptext_)
    , separator(separator_)
{
    if (value_) set(value_, true);
    longest_optarg_name_length = max(get_name().length(), longest_optarg_name_length);
}

// ----------------------------------------------------------------------------------------
//      void RepeatedCmdLineArgumentOption::set(string value_, bool set_default_value)
// ----------------------------------------------------------------------------------------
void RepeatedCmdLineArgumentOption::set(string value_, bool set_default_value) {
    size_t start = 0;
    size_t seppos;
    string next_arg;

    CmdLineOption::set();
    while (string::npos != (seppos = value_.find(separator, start))) {
        next_arg = value_.substr(start, seppos-start);
        if (!next_arg.empty()) RepeatableCmdLineArgument::set(next_arg, set_default_value);
        start    = seppos+1;
    }
    next_arg = value_.substr(start);
    if (!next_arg.empty()) RepeatableCmdLineArgument::set(next_arg, set_default_value);
}

// ---------------------------------------------------------------
//      string RepeatedCmdLineArgumentOption::get_name() const
// ---------------------------------------------------------------
string RepeatedCmdLineArgumentOption::get_name() const {
    const char *Name = CmdLineArgument::get_name().c_str();
    return strf("%s[%c%s]*", Name, separator, Name);
}

// ----------------------------------------------------------------------
//      void RepeatedCmdLineArgumentOption::print(ostream& out) const
// ----------------------------------------------------------------------
void RepeatedCmdLineArgumentOption::print(ostream& out) const
{
    string Name   = get_name();
    int    spaces = longest_optarg_name_length - Name.length() + 1;
    assert(spaces >= 0);
    out << '-' << get_optChar() << Name << string(spaces, ' ');
}

// -end- of implementation of class RepeatedCmdLineArgumentOption.

// start of implementation of class ProgramInfo:

//  --------------------------------------------------------------------------------------------------------------------------------------------------------------
//      ProgramInfo::ProgramInfo(const char *program_name_, int main_version_, int sub_version_, const char *program_description_, const char *creation_date_)
//  --------------------------------------------------------------------------------------------------------------------------------------------------------------
ProgramInfo::ProgramInfo(const char *program_name_, int main_version_, int sub_version_, const char *program_description_, const char *creation_date_)
    : program_name(program_name_),
      main_version(main_version_),
      sub_version(sub_version_),
      program_description(program_description_),
      creation_date(creation_date_),
      any_option_given(false)
{
    Error::set_module_name(program_name_);
}

// --------------------------------------------------------------------------------
//      static bool test_existing_path(const string& candidate,string& result)
// --------------------------------------------------------------------------------
static bool test_existing_path(const string& candidate,string& result) {
    string canonical;
    bool   wrong_path = false;

#if defined(PATH_DETECT)
    cout << "candidate='" << candidate << "'\n";
#endif // PATH_DETECT

    try { canonical = canonicalize(candidate); }
    catch (IOError&) { wrong_path = true; }

#if defined(PATH_DETECT)
    if (!canonical.empty()) cout << "canonical='" << canonical << "'\n";
#endif // PATH_DETECT

    if (!wrong_path &&
        nonnull(canonical.length()) &&
        (fileExists(canonical) || fileExists(canonical+".exe")))
    {
#if defined(PATH_DETECT)
        cout << canonical << " exists!\n";
#endif // PATH_DETECT
        result = extractDirectory(canonical);
#if defined(PATH_DETECT)
        cout << "result='" << result << "'\n";
#endif // PATH_DETECT
        assert(!result.empty());
        return true;
    }

//     if (candidate.find(".exe") == string::npos) {
//         return test_existing_path(candidate+".exe", result);
//     }
    return false;
}

// --------------------------------------------------------------
//      void ProgramInfo::setProgramPaths(const char *argv0)
// --------------------------------------------------------------
void ProgramInfo::setProgramPaths(const char *argv0_) {
    call_path        = getcwd();
    string argv0     = argv0_;

#if defined(PATH_DETECT)
    cout << "argv0='" << argv0 << "'\n";
    cout << "call_path='" << call_path << "'\n";
#endif // PATH_DETECT

    if (!test_existing_path(argv0, program_path) &&
        !test_existing_path(string("./")+argv0, program_path))
    {
        sys_dep::Path path; // reads $PATH from environment
        for (vector<string>::const_iterator dir = path.begin(); dir != path.end(); ++dir) {
            if (test_existing_path(*dir+'/'+argv0_, program_path)) break;
        }
    }

    if (program_path.empty()) {
        throw Error(strf("Can't detect location of %s.exe -- please call with full pathname", program_name.c_str()));
    }

#if defined(PATH_DETECT)
    cout << "program_path='" << program_path << "'\n";
#endif // PATH_DETECT
}
//  ---------------------------------------------------------
//      void ProgramInfo::init(int argc_, char* argv_[])
//  ---------------------------------------------------------
void ProgramInfo::init(int argc_, char* argv_[]) {
    // install terminator
    if (there_can_only_be_one == 0) there_can_only_be_one = new Terminator();

    // test option integrity:
    if (options) {
        bool char_used[256];

        for (int i=0; i<256; ++i) char_used[i] = false;

        for (Options::iterator o = options->begin(); o!=options->end(); ++o) {
            char c = (*o)->get_optChar();

            if (char_used[c]) {
                throw InternalError(string("Duplicated usage of option character '")+c+"'");
            }
            char_used[c] = true;
            if (!options_are_case_sensitive && isalpha(c)) {
                if (islower(c)) {
                    if (char_used[toupper(c)]) options_are_case_sensitive = true;
                }
                else {
                    if (char_used[tolower(c)]) options_are_case_sensitive = true;
                }
            }
        }
    }

    // set defaults for options and arguments
    if (options) {
        for (Options::iterator o=options->begin(); o!=options->end(); ++o) {
            CmdLineArgumentOption *arg_opt = dynamic_cast<CmdLineArgumentOption*>(*o);
            if (arg_opt) { // is an argument option
                arg_opt->set_has_default();
            }
        }
    }

    if (arguments) {
        for (Arguments::iterator a=arguments->begin(); a!=arguments->end(); ++a) {
            (*a)->set_has_default();
        }
    }

    bool allow_multiple_usage = false; // warn if an option is used more than once

    // proceed through command line
    Arguments::iterator next_arg;
    if (arguments) next_arg = arguments->begin();

//     setProgramPaths("once");
    setProgramPaths(argv_[0]);

    for (int a=1; a<argc_; a++) {
        const char *arg = argv_[a];

#if defined(DUMP_PARSE)
        cout << "Parse arg #" << a << ": '" << arg << "'\n";
#endif // DUMP_PARSE

        // test for help-requests
        if ((arg[0]=='/' || arg[0]=='-') &&
            (arg[1]=='?' || arg[1]=='h' || (arg[1]=='-' && strcmp(arg+2, "help")==0)) )
        {
            throw InfoError(*this, false, "Help requested");
        }

        if (arg[0]=='-') {      // option found
            bool internal_option = false;
            if (arg[1] == '-') { // maybe an internal option
                const char *longopt = arg+2;
                internal_option     = true;
                if (rs::str::stricmp(longopt, "multi") == 0) {
                    allow_multiple_usage = true;
                }
                else if (stricmp(string(longopt), "multi") == 0) {
                }
                else {
                    throw StringError(string("Unknown long option '")+arg+"'",
                                      "Valid options are:\n"
                                      "     --help          show command line help\n"
                                      "     --multi         allow to use args multiple (behind this option)\n"
                                      );
                }
            }

            if (!internal_option) {
                size_t pos        = 1;
                bool   done       = false;
                char   switchChar = arg[0];

                while (!done) {
                    char optChar = arg[pos];
                    if (!optChar) {
                        if (pos==1) throw InfoError(*this, string("character awaited behind '") + switchChar + '\'');
                        break;
                    }

                    CmdLineOption *found_option = 0;

                    if (options) {
                        Options::iterator found = find_if(options->begin(), options->end(), HasOptChar(optChar));
                        if (found!=options->end()) {
                            found_option = *found;
                        }
                    }

                    if (!found_option) {
                        throw InfoError(*this, string("Unknown option '") + switchChar + optChar + '\'' );
                    }

                    any_option_given = true;
                    if (found_option->reallyUsed()) { // option was already set
                        if (!allow_multiple_usage) {
                            throw InfoError(*this, string("Option '")+switchChar+optChar+"' was already used (preceed with --multi to avoid this warning).");
                        }
                    }

                    if (found_option->takes_argument()) {
                        CmdLineArgumentOption *arg_option = static_cast<CmdLineArgumentOption*>(found_option);
                        string arg_value;

                        if (arg[pos+1]==0) { // arg_value is in next argv_
                            if (a<(argc_-1)) { // there is another arg
                                arg_value = argv_[++a];
                            }
                        }
                        else { // arg_value appeded to option
                            arg_value = string(arg, pos+1, string::npos);
                        }

                        if (arg_value.empty()) {
                            throw InfoError(*this, string("Option ") + switchChar + optChar + " awaits argument " + arg_option->get_name());
                        }
                        arg_option->set(arg_value);
                        done = true;
                    }
                    else {
                        found_option->set();
                    }
                    pos++;
                }
            }
        }
        else { // it seems to be an argument
            bool arg_handler_found = false; // detected correct arg handler?
            bool inc_arg_handler = true; // continue with next arg handler afterwards

            while (!arg_handler_found) {
                if (!arguments || next_arg==arguments->end()) {
                    throw InfoError(*this, string("Superfluous argument '")+arg+'\'');
                }

                RepeatableCmdLineArgument *repeatable = dynamic_cast<RepeatableCmdLineArgument*>(*next_arg);

                if (repeatable) {
                    inc_arg_handler = false;
                    if (repeatable->get_count() < repeatable->get_min_occur()) { // if we have not scanned enough of this yet
                        arg_handler_found = true; // use it anyway
                    }
                    else { // we have scanned enough of it
                        int max_occur = repeatable->get_max_occur();
                        assert(max_occur==-1 || repeatable->get_count()<max_occur); // otherwise next_arg should have been incremented before
                        string arg_str(arg);
                        if (repeatable->isArg(&arg_str)) { // should be added to repeatable
                            arg_handler_found = true;
                            if (repeatable->get_count()==(max_occur-1)) {
                                inc_arg_handler = true;
                            }
                        }
                        else {
                            ++next_arg; // try next arg
                        }
                    }
                }
                else {
                    arg_handler_found = true;
                }
            }

            (*next_arg)->set(arg);
            if (inc_arg_handler) ++next_arg;
        }
    }

    if (arguments) {
        if (next_arg!=arguments->end()) {
            string err("Missing arguments:");
            bool raise_error = false;

            do {
                RepeatableCmdLineArgument *repeated = dynamic_cast<RepeatableCmdLineArgument*>(*next_arg);
                bool is_missing = true;

                if (repeated) {
                    if (repeated->get_count() >= repeated->get_min_occur()) { // may be enough
                        if (repeated->isArg(0)) { // argument should be missing
                            is_missing = false;
                        }
                    }
                }

                if (is_missing) {
                    err += ' ';
                    err += (*next_arg)->get_name();
                    raise_error = true;
                }
                next_arg++;
            }
            while(next_arg!=arguments->end());

            if (raise_error) throw InfoError(*this, err);
        }
    }

    initialized = true;
}

// --------------------------------------------------------------------------------
//     void ProgramInfo::infos(ostream& out) const
// --------------------------------------------------------------------------------

/** @todo: try to detect terminal width (and print infos according to that width)
 */
void ProgramInfo::infos(ostream& out) const {
    size_t border    = 1;
    size_t tty_width = 80;
    size_t width = tty_width-2*border;
    string left(border, ' ');
    string line  = left+string(width, '-')+'\n';

    out << "\n"
        << line
        << left << program_name << ' ';
    out << int(main_version) << '.' << setw(2) << setfill('0') << int(sub_version);
    out << "   ---   (C) " << creation_date << " by " << COPYRIGHT_BY() << "\n";
    out << line
        << left << "Purpose: "; output::printIndented(out, 9, border, width, program_description);
    out << left << "Usage  : " << program_name;

    if (options) out << " <opts>";

    if (arguments) {
        for (Arguments::iterator i=arguments->begin(); i!=arguments->end(); i++) {
            out << ' ' << (*i)->get_name();
        }
    }
    out << "\n"
        << "\n";

    // show detailed info on arguments:
    if (arguments) {
        for (Arguments::iterator i=arguments->begin(); i!=arguments->end(); i++) {
            string arg_left = (*i)->CmdLineArgument::get_name();
            unsigned len = arg_left.length();

            if (len<longest_arg_name_length) {
                unsigned spaces = longest_arg_name_length-len;
                arg_left += string(spaces, ' ');
                len += spaces;
            }

            arg_left += "::= ";
            len += 4;

            out << left << arg_left; output::printIndented(out, len, border, width, (*i)->get_helptext());
        }
        out << "\n";
    }

    // show detailed info on options:
    string opt_left = "<opts>";
//  unsigned len1 = opt_left.length();

//     if (len1<longest_arg_name_length) {
//         unsigned spaces = longest_arg_name_length-len1;
//         opt_left += string(spaces, ' ');
//         len1 += spaces;
//     }

    opt_left += "::= ";
    //    len1 += 4;
    unsigned len1 = opt_left.length();

    if (!options) options = new Options();
    if (!annotations) annotations = new Annotations();

    Options::iterator i = options->begin();
    Annotations::iterator j = annotations->begin();
    int printed = 0;

    while (i!=options->end()) {
        out << left << opt_left;
        if (i==options->begin()) {
            opt_left = string(len1, ' ');
            assert(opt_left.length()==len1);
        }

        if (j!=annotations->end()) {
            if ((*j)->get_print_infrontof_option()==printed) {
                output::printIndented(out, len1, border, width,
                                      (*j)->print_spaced()
                                      ? string("\n")+(*j)->get_helptext()+"\n"
                                      : (*j)->get_helptext());

                out << left << opt_left;
                ++j;
            }
        }

        (*i)->print(out);
        output::printIndented(out, len1 + 2 + longest_optarg_name_length + 1, border, width, (*i)->get_helptext());

        ++i;
        ++printed;
    }

    if (j!=annotations->end()) {
        opt_left = string(len1, ' ');
        while (j!=annotations->end()) {
            output::printIndented(out, printed ? len1 : 0, border, width, string("\n")+(*j)->get_helptext()+"\n");
            ++j;
        }
    }

    out << line
        << left << "Contact: " << MAIL_CONTACT(program_name) << "\n"
        << line;
}

// --------------------------------------------------------------------------------
//     static string countOptions(CmdLineOptionArray optList, size_t& used_options, const string& trennwort, bool return_used_options)
// --------------------------------------------------------------------------------
static string countOptions(CmdLineOptionArray optList, size_t& used_options, const string& trennwort, bool return_used_options) {
    test_initialized("countOptions");

    used_options = 0;
    unsigned i;

    for (i = 0; optList[i]; i++) {
        if (optList[i]->reallyUsed()) used_options++;
    }

    unsigned size = i;
    size_t counter = 0;
    string opts;
    for (i = 0; optList[i]; i++) {
        if (return_used_options==false || optList[i]->reallyUsed()) { // insert option into return-string?
            counter++;
            bool is_first = counter==1;
            bool is_last = return_used_options ? (counter==used_options) : (counter==size);

            if (is_last) {
                opts = opts+' '+trennwort;
            }
            else if (!is_first) {
                opts += ',';
            }
            opts = opts+" -" + optList[i]->get_optChar();
        }
    }

    return opts;
}

//  --------------------------------------------------------------------------------------------
//      void ProgramInfo::exclude(CmdLineOptionArray optList1, CmdLineOptionArray optList2)
//  --------------------------------------------------------------------------------------------
void ProgramInfo::exclude(CmdLineOptionArray optList1, CmdLineOptionArray optList2) {
    test_initialized("exclude");

    for (unsigned i1=0; optList1[i1]; ++i1) {
        if (optList1[i1]->reallyUsed()) { // used?
            for (unsigned i2=0; optList2[i2]; ++i2) {
                if (optList2[i2]->reallyUsed()) { // used?
                    if (optList1[i1]!=optList2[i2]) { // not same option
                        throw InfoError(*this, string("Option -")+optList1[i1]->get_optChar()+
                                        " can't be used together with -"+optList2[i2]->get_optChar());
                    }
                }
            }
        }
    }
}

// --------------------------------------------------------------------------------
//     void ProgramInfo::needOne(CmdLineOptionArray optList)
// --------------------------------------------------------------------------------
void ProgramInfo::needOne(CmdLineOptionArray optList) {
    size_t Count;
    string ex_opt = countOptions(optList, Count, "or", false);
#if defined(DEBUG)
    cerr << "needOne: '" << ex_opt << "'\n";
#endif // DEBUG
    if (Count==0) {
        string msg = "You have to specify one of the options"+ex_opt+'.';
        throw InfoError(*this, msg);
    }
}

// --------------------------------------------------------------------------------
//     void ProgramInfo::neededBy(const CmdLineOption& drug, const CmdLineOption& addict)
// --------------------------------------------------------------------------------
void ProgramInfo::neededBy(const CmdLineOption& drug, const CmdLineOption& addict) {
    if (addict.reallyUsed() && !drug) {
        throw InfoError(*this, string("Specifying option -")+addict.get_optChar()+ " does not make sense without option -"+drug.get_optChar());
    }
}
//  --------------------------------------------------------------
//      void ProgramInfo::throw_Error(const string& msg) const
//  --------------------------------------------------------------
void ProgramInfo::throw_Error(const string& msg) const {
    throw InfoError(*this, msg);
}

// ----------------------------------------------------
//      const string& ProgramInfo::CallPath() const
// ----------------------------------------------------
const string& ProgramInfo::CallPath() const
{
    test_initialized("CallPath");
    return call_path;
}

// -------------------------------------------------------
//      const string& ProgramInfo::ProgramPath() const
// -------------------------------------------------------
const string& ProgramInfo::ProgramPath() const
{
    test_initialized("ProgramPath");
    return program_path;
}

// -end- of implementation of class ProgramInfo.

