#ifndef OPTION_H_INCLUDED
#define OPTION_H_INCLUDED

#ifndef ERR_H
#include "Err.h"
#endif

#ifndef __LIST__
#include <list>
#endif

namespace rs {

    /** @brief Comfortable command line interface

        @warning You'll have to change MAIL_CONTACT() and COPYRIGHT_BY()

        @par Usage description:

        @li Define a global variable of type ProgramInfo.
        @li Define some CmdLineOption's and CmdLineArgument's.
        @li Call ProgramInfo::init() at start of main.
        @li Afterwards you may call ProgramInfo::exclude(),  ProgramInfo::needOne() and ProgramInfo::neededBy()
        to define the allowed and forbidden combinations of command line switches.

        Now it is ensured that command line was interpreted and all dependencies were verified.

        You may do your own consitency tests afterwards and throw your own InfoError's by calling
        ProgramInfo::throw_Error().

        @par Usage example:

        @include examples/Ex_Option.cpp
        If anything goes wrong, then an InfoError is raised which displays sth like:
        @verbinclude examples/Ex_Option_Output.txt
    */

    namespace cmdline {

        // --------------------------------------------------------------------------------
        //     class CmdLineHelpText
        // --------------------------------------------------------------------------------
        /// Helptext used for several CmdLine..-Items
        class CmdLineHelpText : rs_boost::noncopyable {
            std::string helptext; // helptext printed out by '-?'
        protected:
            /** @brief Create new helptext

                Don't try to create CmdLineHelpText by yourself - it's created from it's child classes.

                @param helptext_    This should be a text describing what the item does and how
                it has to be given on the command line.
            */
            CmdLineHelpText(const char *helptext_) {
                helptext = helptext_;
            }
            virtual ~CmdLineHelpText() {}
        public:
            /// Get the helptext
            virtual std::string get_helptext() const { return helptext; }
        };

        // --------------------------------------------------------------------------------
        //     class CmdLineOption
        // --------------------------------------------------------------------------------
        /// A command line switch without an argument.
        class CmdLineOption : public virtual CmdLineHelpText {
            bool used;          // was option given on command line or set via set()?
            char optChar;       // the option character ('c' for option '-c')
        public:
            /** @brief Declare a new command line option.

                Options should be declared global.

                @param optChar_ The option character (i.e. 'c' for command line option -c).
                The character is case sensitive.

                @param helptext_ A description of the option.
            */
            CmdLineOption(char optChar_, const char *helptext_);
            virtual ~CmdLineOption();

            /** @brief Test an option.
                @return true if the option was given on the command line or set via set(). */
            operator bool() const { return used; }

            /// same as operator bool()
            virtual bool reallyUsed() const { return (bool)*this; }

            /// Manually set the options state to 'given'.
            void set() { used = true; }

            /// Manually set the options state to 'not given'.
            virtual void clear() { used = false; }

            /// get the option character (i.e. 'x' for option -x)
            char get_optChar() const { return optChar; }

        protected:
            friend class ProgramInfo;
            virtual void print(std::ostream& out) const;
            virtual bool takes_argument() const { return false; }
        };

        // --------------------------------------------------------------------------------
        //     class CmdLineArgument
        // --------------------------------------------------------------------------------
        /// A command line argument.
        class CmdLineArgument : public virtual CmdLineHelpText {
            std::string name;   // name of argument (as referred in helptext_)
            std::string value;  // argument given on command line
            bool        has_default_value; // if argument has a default value
            bool        really_used; // if argument was given on command line

        public:
            /** @brief Declare a new command line argument.

                Should be declared global.

                @param name_        The name of the argument as it should be used in the helptext.
                (i.e. "filename" results in <filename>)
                @param helptext_    Help for this command line argument.
                @param value_       Default value (defaults to 0).
                @param is_real_arg  Don't change this (defaults to true)
            */
            CmdLineArgument(const char *name_, const char *helptext_, const char *value_=0, bool is_real_arg=true);
            virtual ~CmdLineArgument();

            bool value_is_set() const { return value.length()>0; }

            /** @brief Test the argument.
                @return true if argument was given on command line, set using set() or has a default value. */
            operator bool() const { return value_is_set(); }

            /** @brief Test the argument.
                @return true if argument was given on command line or set using set(). false if only has a default value.*/
            virtual bool reallyUsed() const { return really_used; }

            /** @brief Get the argument value.
                @return The given command line argument. */
            virtual std::string get() { return value; }

            /** @brief Get the argument value as numeric value (accepts decimal and hexadecimal arguments).
                @return The numeric value of the given command line argument. */
            virtual int getNumeric();

            /** @brief Get the argument value as floating point value.
                @return The floating point value of the given command line argument. */
            virtual double getDouble();

            /** @brief Set the argument value manually.
                @param value_ New value for argument. (Calling set("xxx") has the same effect as if "xxx" has been given on command line.)
                @param set_default_value defaults to false. if true then only the default value is set. */
            virtual void set(std::string value_, bool set_default_value=false) {
                value = value_;
                really_used = !set_default_value;
            }

            /// Clear the argument manually.
            void clear() {
                if (value_is_set()) {
                    value = "";
                    has_default_value = false;
                    really_used = false;
                }
            }

            /// Return name of argument (as shown in helptext: i.e. "<arg>" )
            virtual std::string get_name() const { return std::string("<")+name+'>'; }

        protected:
            friend class ProgramInfo;

            bool has_default() const { return has_default_value; }
            void set_has_default() { if (value_is_set()) has_default_value = true; }

            virtual std::string get_helptext() const {
                if (!has_default()) return CmdLineHelpText::get_helptext();
                return CmdLineHelpText::get_helptext() + " [Default='" + value + "']";
            }
        };

        /** @typedef CmdLineArgumentPredicate
            @brief test if string is the wanted argument

            User-provided function returning true, if the passed string is the wanted argument.

            @note If the passed string pointer is @p NULL, the function should return
            @li true only if argument should be missing
            @li false otherwise (i.e. if further arguments are needed)

            @par Predefined functions:

            @li @ref acceptAnyArgument accepts any argument which is present on command line
        */
        typedef bool (*CmdLineArgumentPredicate)(const std::string *arg);

        //  -----------------------------------------------------------------
        //      class RepeatableCmdLineArgument : public CmdLineArgument
        //  -----------------------------------------------------------------
        /** @brief base class for optional and/or repeated command line arguments
            @warning Do NOT use this class (use child classes instead)!!!
        */
        class RepeatableCmdLineArgument : public CmdLineArgument {
        private:
            int                      min_occur;
            int                      max_occur;
            std::list<std::string>   repeated_values;
            std::list<std::string>   seen_values;
            CmdLineArgumentPredicate isArg;
            int                      count; // arguments scanned

        protected:
            /// RepeatableCmdLineArgument may only be constructed by it's child classes
            RepeatableCmdLineArgument(const char *name_, const char *helptext_, int min_occur_, int max_occur_, CmdLineArgumentPredicate isArg_, bool is_real_arg);
        public:
            virtual ~RepeatableCmdLineArgument() {}

            /** @brief Get next of the repeated command line arguments.

                After all values have been read via this function, operator bool()
                returns false. Further calls return an empty string.

                @return next command line argument.
            */
            std::string get();

            /** @brief Manually add new argument.
                @param value_ Value of the argument.
                @param set_default_value defaults to false.

                if true then reallyUsed() remains false. Use this to manually add an
                RepeatableCmdLineArgument without setting the reallyUsed-flag.
            */
            virtual void set(std::string value_, bool set_default_value=false) {
                if (value_is_set()) repeated_values.push_back(value_);
                else CmdLineArgument::set(value_, set_default_value);
                ++count;
            }

            /// resets to the first argument
            void reset();

            /// Return name of argument (as shown in helptext: i.e. "[<arg>]+" )
            virtual std::string get_name() const;

        protected:
            friend class ProgramInfo;

            int get_min_occur() const { return min_occur; }
            int get_max_occur() const { return max_occur; }
            int get_count() const { return count; }
        };

        /** accepts every argument on command line
            @see RepeatedCmdLineArgument
        */
        bool acceptAnyArgument(const std::string *arg);

        //  -------------------------------------------------------------------------
        //      class OptionalCmdLineArgument : public RepeatableCmdLineArgument
        //  -------------------------------------------------------------------------
        /// An optional and repeatable command line argument
        class OptionalCmdLineArgument : public RepeatableCmdLineArgument {
        public:
            /** @brief Declare a optional CmdLineArgument

                Should be declared global.

                @param name_        The name of the argument as it should be used in the helptext.
                (i.e. "filename" results in <filename>)
                @param helptext_    Help for this command line argument.
                @param isArg_       test if found argument is the optional argument
                Default: acceptAnyArgument(), which can be used to accept
                every argument which is present on the command line.
                @param max_occur_   maximum repeat count for this type of argument (defaults to 1).
                set to -1 to allow 'infinite' arguments.
            */
            OptionalCmdLineArgument(const char *name_, const char *helptext_, CmdLineArgumentPredicate isArg_ = acceptAnyArgument, int max_occur_=1)
                : RepeatableCmdLineArgument(name_, helptext_, 0, max_occur_, isArg_, true)
                , CmdLineHelpText(helptext_)
            {}
            virtual ~OptionalCmdLineArgument() {}
        };

        //  -------------------------------------------------------------------------
        //      class RepeatedCmdLineArgument : public RepeatableCmdLineArgument
        //  -------------------------------------------------------------------------
        /// A repeatable command line argument.
        class RepeatedCmdLineArgument : public RepeatableCmdLineArgument {
        public:
            /** @brief Declare a new repeatable command line argument.

                RepeatedCmdLineArguments should be declared global.

                @param name_        The name of the argument as it should be used in the helptext.
                (i.e. "filename" results in [<filename>]+)
                @param helptext_    Help for this command line argument.
                @param isArg_       user provided function returning true, if passed string is the optional argument.
                If the passed string is NULL the function should return true only if argument should be missing.
                @param min_occur_   The argument has to be mentioned min_occur_ times on command line (defaults to 1)
                @param max_occur_   The argument has to be mentioned max_occur_ times on command line (defaults to -1 = 'infinite')
                line.
                @see acceptAnyArgument
            */
            RepeatedCmdLineArgument(const char *name_, const char *helptext_, CmdLineArgumentPredicate isArg_, int min_occur_=1, int max_occur_=-1)
                : RepeatableCmdLineArgument(name_, helptext_, min_occur_, max_occur_, isArg_, true)
                , CmdLineHelpText(helptext_)
            {}

            virtual ~RepeatedCmdLineArgument() {}
        };


        // --------------------------------------------------------------------------------
        //     class CmdLineArgumentOption : public CmdLineOption, public CmdLineArgument
        // --------------------------------------------------------------------------------
        /// A command line option which has to be followed by an argument.
        class CmdLineArgumentOption : public CmdLineOption, public CmdLineArgument {
        public:
            /** @brief Declare a new argument option.

                CmdLineArgumentOptions should be declared global.

                @param optChar_     The option character (i.e. 'c' for command line option -c).
                The character is case sensitive.
                @param argname_     The name of the argument as it should be used in the helptext.
                (i.e. "filename" results in <filename>)
                @param helptext_    A description of the option.
                @param value_       Default value (defaults to 0). If value_ is set then operator bool returns true.
            */
            CmdLineArgumentOption(char optChar_, const char *argname_, const char *helptext_, const char *value_=0);
            virtual ~CmdLineArgumentOption();

            /** @brief Manually set the options state to 'given' and set the value of the argument of the option.
                @param value_               the new value
                @param set_default_value    if true -> reallyUsed() returns false
            */
            virtual void set(std::string value_, bool set_default_value=false) {
                CmdLineOption::set();
                CmdLineArgument::set(value_, set_default_value); // set default argument
            }

            /// Manually set the state of the option to 'not given'.
            virtual void clear() {
                CmdLineOption::clear();
                CmdLineArgument::clear();
            }

            /** @brief Return true only if option was given on command line or set using set.

                operator bool() also returns true if the option was set by specifying a default value.
            */
            virtual bool reallyUsed() const { return CmdLineArgument::reallyUsed(); }

        protected:
            virtual void print(std::ostream& out) const;
            virtual bool takes_argument() const { return true; }
//             virtual string get_helptext() const { return CmdLineHelpText::get_helptext(); }
        };

        // ----------------------------------------------------------------------------------------------------
        //      class RepeatedCmdLineArgumentOption
        // ----------------------------------------------------------------------------------------------------
        /** @brief an option with a repeatable argument

            Example: -e "text1;text2;text3"
        */
        class RepeatedCmdLineArgumentOption : public CmdLineOption, public RepeatableCmdLineArgument {
        private:
            char separator;

        public:
            /** @brief declare a command line switch with repeated argument
                @param optChar_     The option character (i.e. 'c' for command line option -c). The character is case sensitive.
                @param argname_     The name of the argument as it should be used in the helptext. (i.e. "filename" results in <filename>)
                @param separator_   The separator for the argument
                @param helptext_    A description of the option.
                @param value_       Default value (defaults to 0). If value_ is set then operator bool returns true.
            */
            RepeatedCmdLineArgumentOption(char optChar_, const char *argname_, char separator_, const char *helptext_, const char *value_=0);
            virtual ~RepeatedCmdLineArgumentOption() {}

            /** @brief Manually set the options state to 'given' and set the value of the argument of the option.
                @param value_               the new value (including the separator's)
                @param set_default_value    if true -> reallyUsed() returns false
            */
            virtual void set(std::string value_, bool set_default_value=false);

            /** @brief Get the next part of the separated arguments

                After all values have been read via this function, operator bool()
                returns false. Further calls return an empty string.

                @return next part
            */
            std::string get() { return RepeatableCmdLineArgument::get(); }

            /// Manually set the state of the option to 'not given'.
            virtual void clear() {
                CmdLineOption::clear();
                RepeatableCmdLineArgument::clear();
            }

            /// Return name of argument option (as shown in helptext: i.e. '<arg>[,<arg>]*' )
            virtual std::string get_name() const;

            /** @brief Return true only if option was given on command line or set using set.

                operator bool() also returns true if the option was set by specifying a default value.
            */
            virtual bool reallyUsed() const { return CmdLineArgument::reallyUsed(); }

        protected:
            virtual void print(std::ostream& out) const;
            virtual bool takes_argument() const { return true; }
//             virtual string get_helptext() const { return CmdLineHelpText::get_helptext(); }
        };


        /// NULL-terminated C-Array of pointers to CmdLineOption
        typedef CmdLineOption** CmdLineOptionArray;


        // --------------------------------------------------------------------------------
        //     class CmdLineAnnotation
        // --------------------------------------------------------------------------------
        /// A annotation for some command line arguments.
        class CmdLineAnnotation : public virtual CmdLineHelpText {
        private:
            int print_infrontof_option;
            bool spaced; // print empty lines above and below
        public:
            /** @brief Declare a new annotation for command line options.

                The annotation will be printed inbetween the two options it is declared inbetween.

                Annotations are printed over the full width of the current block. Use them for more
                detailed descriptions of how sth works.

                @param helptext_ The text of the annotation.
                @param spaced_ Insert empty lines before and after the CmdLineAnnotation
            */
            CmdLineAnnotation(const char *helptext_="", bool spaced_=false);
            virtual ~CmdLineAnnotation() {}

        protected:
            friend class ProgramInfo;
            int get_print_infrontof_option() const { return print_infrontof_option; }
            bool print_spaced() const { return spaced; }
        };


        // --------------------------------------------------------------------------------
        //     class ProgramInfo
        // --------------------------------------------------------------------------------
        /** @brief Global program information
            @see rs::cmdline for an overview how to use ProgramInfo
         */
        class ProgramInfo : rs_boost::singleton<ProgramInfo> {
            std::string program_name;
            int         main_version;
            int         sub_version;
            std::string program_description;
            std::string creation_date;

            bool any_option_given;

            void setProgramPaths(const char *argv0);

            std::string program_path; // home directory of the program
            std::string call_path; // the directory the program was called from

        public:
            /** @brief Create the global program information.

                Every program using the command line
                interface has to declare exactly one global instance of ProgramInfo.

                @param  program_name_           The name of the program.
                @param  main_version_           The main version of the program.
                @param  sub_version_            The sub version of the program.
                @param  program_description_    A description of the purpose of the program.
                @param  creation_date_          Date of creation (defaults to NOW).

                @note When an instance of ProgramInfo exists, all errors will print sth like
                'Error in program_name_: blabla'.
                This is done by defining an err::ErrorModule with name 'program_name_'.

                @see init() has to be called by main
            */
            ProgramInfo(const char *program_name_, int main_version_, int sub_version_,
                        const char *program_description_, const char *creation_date_ = __DATE__);
            ~ProgramInfo() {}

            /** @brief Process command line options and arguments.

                When init is called, all declared options and arguments are scanned.

                If option -?, -h or --help is given or if an error occurs, then some help
                for the usage of the program is displayed.

                @param argc_ same as argc in main
                @param argv_ same as argv in main
            */
            void init(int argc_, char* argv_[]); // do not change!

            /** @name Exclude
                @anchor Exclude
                @brief Declaration of options that cannot be used together.

                Each of the exclude-functions declares two (sets of) CmdLineOption's
                which may not be used together. If the user of your program uses any of
                them together, an InfoError will be raised.
            */
            //@{
            /** @param opt1 Reference to (or array of) CmdLineOption's
                @param opt2 Reference to (or array of) CmdLineOption's

                @note None of the options in @p opt1 may be used together
                with any from @p opt2.

                @ref Exclude "General info"
            */

            void exclude(CmdLineOptionArray opt1, CmdLineOptionArray opt2);
            void exclude(CmdLineOption& opt1, CmdLineOptionArray opt2) {
                CmdLineOption *optList1[] = { &opt1, 0 };
                exclude(optList1, opt2);
            }
            void exclude(CmdLineOptionArray opt1, CmdLineOption& opt2) {
                CmdLineOption *optList2[] = { &opt2, 0 };
                exclude(opt1, optList2);
            }
            void exclude(CmdLineOption& opt1, CmdLineOption& opt2) {
                CmdLineOption *optList1[] = { &opt1, 0 };
                CmdLineOption *optList2[] = { &opt2, 0 };
                exclude(optList1, optList2);
            }

            //@}

            /** @brief Declare options, where one of them has to be given.
                @param options Array of options, where one has to be given. */
            void needOne(CmdLineOptionArray options);

            /** @brief Declare dependencies between CmdLineOptions.
                @param drug needed by addict
                @param addict needs the drug */
            void neededBy(const CmdLineOption& drug, const CmdLineOption& addict);

            /** @brief Display program help.
                @param out where to print */
            void infos(std::ostream& out) const;

            /** @brief Test if any options were given on command line
                @return true if options were given. */
            bool optionsSpecified() const { return any_option_given; }

            /** @brief Throws an InfoError. If you print the error the program help is also shown.
                @param msg the error message
                @note InfoError is a hidden class derived from @ref err::Error
            */
            void throw_Error(const std::string& msg) const;

            /// returns the directory the program was called from
            const std::string& CallPath() const;

            /// returns the directory where the program is stored
            const std::string& ProgramPath() const;
        };

    };
};

#else
#error Option.h included twice
#endif

