/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000
// Ralf Westram
// (Coded@ReallySoft.de)
//
// Permission to use, copy, modify, distribute and sell this software
// and its documentation for any purpose is hereby granted without fee,
// provided that the above copyright notice appear in all copies and
// that both that copyright notice and this permission notice appear
// in supporting documentation.  Ralf Westram makes no
// representations about the suitability of this software for any
// purpose.  It is provided "as is" without express or implied warranty.
//
// This code is part of my library.
// You may find a more recent version at http://www.reallysoft.de/
//
/////////////////////////////////////////////////////////////////////////////

#ifndef TIMER_H
#define TIMER_H

#include <ctime>
#include <string>
#include <iostream>

#ifdef NDEBUG
// define SHOW_TIMERS before including header to show Timers in non-debug code
#else
# ifndef DONT_SHOW_TIMERS
// define DONT_SHOW_TIMERS before including header to hide Timers in debug code
#  define SHOW_TIMERS
# endif
#endif

namespace rs {

    // --------------------------------------------------------------------------------
    //     class Timer
    // --------------------------------------------------------------------------------

    /** A simple timer class (just working with seconds)
     */
    class Timer {
    private:
        time_t        start;
        std::string   description;
        std::ostream& out;
        size_t        counter;
        std::string   count_description;
        static int    timer_counter;

        size_t uptime() {
            time_t now;
            time(&now);
            return now-start;
        }

        void indent() {
            for (int i=0; i<timer_counter; i++) {
                out << '*';
            }
        }

    public:
        /** start timer

            Timer prints its info when it is destroyed.
            So all you have to do is declare a timer in an appropriate block.

            @param aOut where to print timer info
            @param desc description printed with timer
        */
        Timer(std::ostream& aOut, const std::string& desc)
            : description(desc),
              out(aOut),
              counter(0),
              count_description("none")
        {
            time(&start);
            timer_counter++;
#if !defined(SHOW_TIMERS)
            indent();
            out << "Started " << description << "..\n";
#endif // SHOW_TIMERS
        }

        void show() {
#if !defined(SHOW_TIMERS)
            size_t seconds = uptime();
            if (seconds==0) seconds = 1;

            indent();
            out << "Stopped " << description << " after " << seconds << " seconds.";
            if (counter) {
                out << ' ' << count_description << '=' << counter << "; per sec=" << double(counter)/seconds;
            }
            out << "\n";
#endif // SHOW_TIMERS
        }

        virtual ~Timer() {
            show();
            timer_counter--;
        }

        /** set a counter for a timer

            if you are timing single actions and for example do a loop over 10000 single actions
            you should set the counter to 10000. If you do the mean time for a single action is
            calculated.

            @param aCount       number of single actions
            @param count_desc   description of single actions
         */
        void set_counter(size_t aCount, const std::string& count_desc) {
            counter = aCount;
            count_description = count_desc;
        }
    };

};

#else
#error Timer.h included twice
#endif // TIMER_H
