//  ======================================================================== //
//                                                                           //
//    File      : boost_util.h                                               //
//    Purpose   : Code from boost library (see http://www.boost.org/)        //
//    Time-stamp: <Wed Nov/13/2002 08:30 MET Coder@ReallySoft.de>            //
//                                                                           //
//    (C) May 2002 by Ralf Westram                                           //
//                                                                           //
//    Permission to use, copy, modify, distribute and sell this software     //
//    and its documentation for any purpose is hereby granted without fee,   //
//    provided that the above copyright notice appear in all copies and      //
//    that both that copyright notice and this permission notice appear      //
//    in supporting documentation.                                           //
//                                                                           //
//    Ralf Westram makes no representations about the suitability of this    //
//    software for any purpose.  It is provided "as is" without express or   //
//    implied warranty.                                                      //
//                                                                           //
//  ======================================================================== //

#ifndef BOOST_UTIL_H
#define BOOST_UTIL_H

#ifndef CSTDLIB
#include <cstdlib>
#endif


namespace rs {

    /** @brief code from the boost-library (plus some similar additional code)

        see http://www.boost.org/ for the complete boost library
     */
    namespace rs_boost {

        /** @brief next() and prior() template functions

            Helper functions for classes like bidirectional iterators not supporting
            operator+ and operator-.

            @par Usage:

            @code
            const std::list<T>::iterator p    = get_some_iterator();
            const std::list<T>::iterator prev = boost::prior(p);
            @endcode

            Contributed by Dave Abrahams
        */

        template <class T>
        T next(T x) { return ++x; }

        /// opposite of @ref next
        template <class T>
        T prior(T x) { return --x; }


        /** @brief make derived classes uncopyable

            Private copy constructor and copy assignment ensure classes derived from
            class noncopyable cannot be copied.

            Contributed by Dave Abrahams
        */

        class noncopyable
        {
        protected:
            noncopyable(){}
            ~noncopyable(){}
        private:  // emphasize the following members are private
            noncopyable( const noncopyable& );
            const noncopyable& operator=( const noncopyable& );
        };

        /** @brief counts the number of instances

            When a class X is derived from counted<X> the current and maximum number
            of instances of X (and all classes derived from X) can be determined
        */
        template <class X>
        class counted {
            static size_t instances;
            static size_t max_instances;
        public:
            counted() {
                ++instances;
                if (instances>max_instances) max_instances = instances;
            }
            ~counted() { --instances; }

            size_t get_instances() const { return instances; }
            size_t get_max_instances() const { return max_instances; }
        };

        template <class X> size_t counted<X>::instances = 0;
        template <class X> size_t counted<X>::max_instances = 0;

        void second_instance_error();

        /** @brief ensure that only one object of a class may be constructed

            Simply derive your class from singleton.

            @code
            class MyClass : singleton<MyClass> { ... };
            @endcode
        */
        template <class X>
        class singleton : counted<X> {
        public:
            /// throws @ref err::InternalError when a second instance is constructed
            singleton() : counted<X>()  {
                if (get_instances()>1) second_instance_error();
            }
            ~singleton() {}
        };
    };

}; // end of namespace rs

#else
#error boost_util.h included twice
#endif // BOOST_UTIL_H

