/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000
// Ralf Westram
// (Coded@ReallySoft.de)
//
// Permission to use, copy, modify, distribute and sell this software
// and its documentation for any purpose is hereby granted without fee,
// provided that the above copyright notice appear in all copies and
// that both that copyright notice and this permission notice appear
// in supporting documentation.  Ralf Westram makes no
// representations about the suitability of this software for any
// purpose.  It is provided "as is" without express or implied warranty.
//
// This code is part of my library.
// You may find a more recent version at http://www.reallysoft.de/
//
/////////////////////////////////////////////////////////////////////////////

#ifndef DBF_H
#define DBF_H

#ifndef __CTIME__
#include <ctime>
#endif
#ifndef __FSTREAM__
#include <fstream>
#endif
#ifndef __MAP__
#include <map>
#endif


#ifndef ERR_H
#include "Err.h"
#endif
#ifndef SMARTPTR_H
#include "smartptr.h"
#endif


namespace rs {

    /// DBase compatible database reader
    namespace dbf {

        // --------------------------------------------------------------------------------
        //     class DBF_EntryDeclaration
        // --------------------------------------------------------------------------------

        /// Info about one Dbase entry type
        class DBF_EntryDeclaration {
        private:
            std::string name;
            char type;
            char xtra[4];
            unsigned char fieldlen;
            char right;
            unsigned offset; // in record
            unsigned index; // index of field in record

        public:
            DBF_EntryDeclaration(unsigned char firstChar, std::istream& in, unsigned& offset_in_record, unsigned& index_);
            virtual ~DBF_EntryDeclaration() {}

            const std::string& Name() const { return name; }
            unsigned Fieldlen() const { return fieldlen; }
            unsigned Offset() const { return offset; }
            unsigned Index() const { return index; }

            bool is_numeric() const { return type=='N'; }
            bool is_string() const { return type=='C'; }

            void dump(std::ostream& out) const;
            DECLARE_STREAM_OP(DBF_EntryDeclaration);

            bool operator<(const DBF_EntryDeclaration& other) const { return name<other.name; }
        };

        typedef SmartPtr<DBF_EntryDeclaration> DBF_EntryDeclarationPtr;
        typedef std::map<std::string, DBF_EntryDeclarationPtr> DBF_EntryDeclarationList;

        // --------------------------------------------------
        //      class DBF_Reader : rs_boost::noncopyable
        // --------------------------------------------------
        /// DBase-compatible database reader class
        class DBF_Reader : rs_boost::noncopyable {
        private:
            std::string filename;
            std::ifstream in;

            struct tm created;
            unsigned no_of_records;
            unsigned length_of_header;
            unsigned record_size;

            DBF_EntryDeclarationList entryTypes;

            unsigned current_record;
            char *current_record_data;

            void read_record(unsigned record_number) {
                if (record_number!=current_record) {
                    in.seekg(length_of_header+record_number*record_size+1);
                    in.read(current_record_data, record_size);
                    current_record = record_number;
                }
            }

        public:
            /** Start reading an existing database
                @param filename_ name of database %file

                The record pointer is positioned onto the first record.

                @see gotoRecord
                @see nextRecord

                @exception rs::Error if %file not found, wrong filetype or database empty.
            */
            DBF_Reader(const std::string& filename_);
            virtual ~DBF_Reader() { delete current_record_data; }

            /** Dump entries defined in database (use this to detect which fields are declared in a database)
                @param out output stream to which the entries will be dumped */
            void dumpEntryDefinition(std::ostream& out) const;

            /** Get number of fields in each record
                @return number of fields
            */
            unsigned Fields() const { return entryTypes.size(); }

            /** Get name of field
                @param num number of field (range: 0..(Fields()-1))
                @return string containing name of field
                @exception Error if index out of range
            */
            const std::string& FieldName(unsigned num) const;

            /** Get number of records stored in database
                @return number of records */
            unsigned Records() const { return no_of_records; }

            /** Proceed to next record
                @return true if successful */
            bool nextRecord() {
                if (current_record==no_of_records-1) return false;
                read_record(current_record+1);
                return true;
            }

            /** Goto specific record
                @param index index of record (range: 0..(Records()-1))
                @return true if successful */
            bool gotoRecord(unsigned index) {
                if (index>=no_of_records) return false;
                read_record(index);
                return true;
            }

            /** Read content of current record
                @param fieldName name of field to read
                @see FieldName (to get legal values for fieldName)
                @param crop if true -> crop leading and trailing spaces from result (defaults to true)
                @return string containing content
                @exception Error if fieldName is illegal
            */
            std::string FieldContent(const std::string& fieldName, bool crop=true) const;
        };
    };
};

#else
#error dbf.h included twice
#endif // DBF_H
