/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000
// Ralf Westram
// (Coded@ReallySoft.de)
//
// Permission to use, copy, modify, distribute and sell this software
// and its documentation for any purpose is hereby granted without fee,
// provided that the above copyright notice appear in all copies and
// that both that copyright notice and this permission notice appear
// in supporting documentation.  Ralf Westram makes no
// representations about the suitability of this software for any
// purpose.  It is provided "as is" without express or implied warranty.
//
// This code is part of my library.
// You may find a more recent version at http://www.reallysoft.de/
//
/////////////////////////////////////////////////////////////////////////////

#include "inifile.h"
#include "InputFile.h"

#ifndef __FSTREAM__
#include <fstream>
#endif

using namespace std;
using namespace rs;
using namespace rs::err;
using namespace rs::file;
using namespace rs::ini;
using namespace rs::str;

namespace rs {
    namespace ini {

        bool need_check_for_changes = false;

    };
};

//  ----------------------------------------
//      void Comment::split_multiline()
//  ----------------------------------------
void Comment::split_multiline() {
    size_t cr_pos = text.find('\n');
    while (cr_pos!=string::npos) {
        text.insert(cr_pos+1, 1, comment_char);
        cr_pos = text.find('\n', cr_pos+1);
    }
}
//  -----------------------------------------------
//      void Comment::append(Comment *comment)
//  -----------------------------------------------
void Comment::append(Comment *comment) {
    Comment *cmt = this;
    while (cmt->next) cmt = cmt->next;
    assert(cmt->next==0);
    assert(comment->next==0);
    cmt->next = comment;
}
//  -----------------------------------------------
//      void Comment::dump(ostream& out) const
//  -----------------------------------------------
void Comment::dump(ostream& out) const {
    const Comment *cmt = this;
    while (cmt) {
        out << comment_char << cmt->text << '\n';
        cmt = cmt->next;
    }
}

// start of implementation of class Entry:

// ------------------------------------------------
//      double Entry::getNumericContent() const
// ------------------------------------------------
double Entry::getNumericContent() const
{
    const char *c       = content.c_str();
    const char *e       = 0;
    double      value   = strtod(c, const_cast<char**>(&e));

    if (e && e[0]) { // sth behind value
        while (e[0] == ' ') ++e; // skip spaces
        if (e[0]) throw_Error(strf("Unexpected content (%s) found behind/instead of numeric value", e));
    }

    return value;
}

// ----------------------------------------------
//      long Entry::getIntegerContent() const
// ----------------------------------------------
long Entry::getIntegerContent() const
{
    const char *c     = content.c_str();
    const char *e     = 0;
    long        value = strtol(c, &const_cast<char*>(e), 10);

    if (e && e[0]) { // sth behind value
        while (e[0] == ' ') ++e; // skip spaces
        if (e[0]) throw_Error(strf("Unexpected content (%s) found behind/instead of numeric value", e));
    }

    return value;
}


// --------------------------------------------------------------
//      void Entry::throw_Error(const string& message) const
// --------------------------------------------------------------
void Entry::throw_Error(const string& message) const
{
    string mess2 = strf("In [%s]/%s: %s", mySection()->getName().c_str(), getName().c_str(), message.c_str());
    FilePosition::throw_Error(mySection()->myInifile()->loadedFrom(), mess2);
}

// -end- of implementation of class Entry.

// start of implementation of class Section:


// ---------------------------------------------------------------------------------------------------------------------------------
//      EntryPtr Section::findOrCreateEntry(const string& name_, long default_value, ChangesState mode, const char *add_comment)
// ---------------------------------------------------------------------------------------------------------------------------------
EntryPtr Section::findOrCreateEntry(const string& name_, long default_value, ChangesState mode, const char *add_comment)
{
    EntryPtr entry = findEntry(name_);
    if (entry.Null()) {
        string default_string_value = strf("%li", default_value);
        insertEntry(new Entry(name_, default_string_value, UNDEFINED_FILEPOSITION), mode);
        entry                       = findEntry(name_);
        if (add_comment) entry->append(add_comment);
    }
    return entry;
}

//  --------------------------------------------------
//      ChangesState Section::changesMade() const
//  --------------------------------------------------
ChangesState Section::changesMade() const {
    if (changes_made == NO_CHANGES && need_check_for_changes)  {
        for (Entries::const_iterator e = begin(); changes_made == NO_CHANGES && e != end(); ++e) {
            if (e->second->changesMade()) changes_made = MINOR_CHANGES;
        }
    }

    return changes_made;
}

// ------------------------------------------------------------------------------------------------------------------------------------
//      EntryPtr Section::findOrCreateEntry(const string& name_, const string& content, ChangesState mode, const char *add_comment)
// ------------------------------------------------------------------------------------------------------------------------------------
EntryPtr Section::findOrCreateEntry(const string& name_, const string& content, ChangesState mode, const char *add_comment)
{
    EntryPtr entry = findEntry(name_);
    if (entry.Null()) {
        insertEntry(new Entry(name_, content, UNDEFINED_FILEPOSITION), mode);
        entry = findEntry(name_);
        if (add_comment) entry->append(add_comment);
    }
    return entry;
}

// --------------------------------------------------------------------------------------------------------------------------------------
//      EntryPtr Section::findOrCreateEntry(const string& name_, string (*get_content)(), ChangesState mode, const char *add_comment)
// --------------------------------------------------------------------------------------------------------------------------------------
EntryPtr Section::findOrCreateEntry(const string& name_, string (*get_content)(), ChangesState mode, const char *add_comment)
{
    EntryPtr entry = findEntry(name_);
    if (entry.Null()) {
        insertEntry(new Entry(name_, get_content(), UNDEFINED_FILEPOSITION), mode);
        entry = findEntry(name_);
        if (add_comment) entry->append(add_comment);
    }
    return entry;
}

// --------------------------------------------------------------------------
//      const string& Section::findStringEntry(const string& name_) const
// --------------------------------------------------------------------------
const string& Section::findStringEntry(const string& name_) const
{
    return expectEntry(name_)->getContent();
}

// --------------------------------------------------------------------
//      double Section::findNumericEntry(const string& name_) const
// --------------------------------------------------------------------
double Section::findNumericEntry(const string& name_) const
{
    return expectEntry(name_)->getNumericContent();
}

// -------------------------------------------------------------------
//      long Section::findIntegerEntry(const string& name_) const
// -------------------------------------------------------------------
long Section::findIntegerEntry(const string& name_) const {
    return expectEntry(name_)->getIntegerContent();
}

// ---------------------------------------------------------------
//      void Section::throw_Error(const string& message) const
// ---------------------------------------------------------------
void Section::throw_Error(const string& message) const
{
    FilePosition::throw_Error(myInifile()->loadedFrom(), strf("In section [%s]: %s", getName().c_str(), message.c_str()));
}

// -----------------------------------------------------------------
//      EntryPtr Section::expectEntry(const string& name_) const
// -----------------------------------------------------------------
EntryPtr Section::expectEntry(const string& name_) const
{
    EntryPtr entry = findEntry(name_);
    if (entry.Null()) throw_Error(string("No such entry: '")+name_+'\'');
    return entry;
}

// -end- of implementation of class Section.

// start of implementation of class Inifile:

// ----------------------------------------------------------------------------------------
//      SectionPtr Inifile::findOrCreateSection(const string& name_, ChangesState mode)
// ----------------------------------------------------------------------------------------
SectionPtr Inifile::findOrCreateSection(const string& name_, ChangesState mode)
{
    SectionPtr section = findSection(name_);
    if (section.Null()) {
        insertSection(new Section(name_, UNDEFINED_FILEPOSITION), mode);
        section = findSection(name_);
        assert(!section.Null());
    }
    return section;
}

//  --------------------------------------------------
//      ChangesState Inifile::changesMade() const
//  --------------------------------------------------
ChangesState Inifile::changesMade() const {
    for (Sections::const_iterator s = sections.begin(); changes_made!=MAJOR_CHANGES && s!=sections.end(); ++s) {
        ChangesState section_state = s->second->changesMade();
        if (section_state>changes_made) changes_made = section_state;
    }
    return changes_made;
}
//  ---------------------------------
//      void Inifile::fixState()
//  ---------------------------------
void Inifile::fixState() {
    for (Sections::iterator s = sections.begin(); s!=sections.end(); ++s) {
        s->second->fixState();
    }
    changes_made = NO_CHANGES;
}

// -------------------------------------------------------------------------------------------------
//      const string& Inifile::findStringEntry(const string& section, const string& entry) const
// -------------------------------------------------------------------------------------------------
const string& Inifile::findStringEntry(const string& section, const string& entry) const
{
    return expectSection(section)->findStringEntry(entry);
}

// -------------------------------------------------------------------------------------------
//      double Inifile::findNumericEntry(const string& section, const string& entry) const
// -------------------------------------------------------------------------------------------
double Inifile::findNumericEntry(const string& section, const string& entry) const
{
    return expectSection(section)->findNumericEntry(entry);
}
// -----------------------------------------------------------------------------------------
//      long Inifile::findIntegerEntry(const string& section, const string& entry) const
// -----------------------------------------------------------------------------------------
long Inifile::findIntegerEntry(const string& section, const string& entry) const
{
    return expectSection(section)->findIntegerEntry(entry);
}

// --------------------------------------------------------------------------------------------
//      const EntryPtr Inifile::findEntry(const string& section, const string& entry) const
// --------------------------------------------------------------------------------------------
const EntryPtr Inifile::findEntry(const string& section, const string& entry) const
{
    return expectSection(section)->expectEntry(entry);
}

// ----------------------------------------------------------------------
//      SectionPtr Inifile::expectSection(const string& name_) const
// ----------------------------------------------------------------------
SectionPtr Inifile::expectSection(const string& name_) const
{
    SectionPtr s = findSection(name_);
    if (s.Null()) throw Error(strf("In %s: No such section: [%s]", loadedFrom().c_str(), name_.c_str()));
    return s;
}

// -------------------------------------------------------------
//      SectionPtr Inifile::findSection(const string& name_)
// -------------------------------------------------------------
SectionPtr Inifile::findSection(const string& name_)
{
    Sections::iterator s = sections.find(name_);
    if (s==sections.end()) return SectionPtr();
    return s->second;
}

// -end- of implementation of class Inifile.

// start of implementation of class FilePosition:

// --------------------------------------------------------------------------------------------
//      void FilePosition::throw_Error(const string& filename, const string& message) const
// --------------------------------------------------------------------------------------------
void FilePosition::throw_Error(const string& filename, const string& message) const
{
    throw Error(strf("In %s (#%lu): %s", filename.c_str(), lineNumber, message.c_str()));
}

// -end- of implementation of class FilePosition.


namespace rs {
    namespace ini {
        // --------------------------------------------------------------------------------
        //     InifilePtr readInifile(const string& filename)
        // --------------------------------------------------------------------------------
        InifilePtr readInifile(const string& filename) {
            InifilePtr ini(new Inifile(filename));
            SectionPtr currentSection;
            EntryPtr currentEntry;

            InputFile in(filename);
            string line;
            Comment *comment = 0;

            while (in.getLine(line)) {
                const char *linestart = line.c_str();
                while (linestart[0]==' ') ++linestart;

                switch (linestart[0]) {
                    case 0: break; // eol

                        // comments
                    case ';':
                    case '#': {
                        if (comment) comment->append(linestart+1);
                        else comment = new Comment(linestart+1);
                        break;
                    }

                    case '[': { // section
                        const char *kl_zu = strchr(linestart, ']');
                        if (!kl_zu) in.throw_Error("Missing ']'");

                        string sectionName(linestart+1, 0, kl_zu-linestart-1);
                        currentSection = new Section(sectionName, in.getLineNumber());
                        if (comment) {
                            currentSection->append(comment);
                            comment = 0;
                        }
                        ini->insertSection(currentSection, NO_CHANGES);
                        break;
                    }
                    default: { // entry
                        const char *gleich = strchr(linestart, '=');
                        if (!gleich) in.throw_Error("Missing '='");

                        if (currentSection.Null()) in.throw_Error(string("Entry without section"));

                        string entryName(linestart, 0, gleich-linestart);
                        string content(gleich+1);

                        currentEntry = new Entry(entryName, content, in.getLineNumber());
                        if (comment) {
                            currentEntry->append(comment);
                            comment = 0;
                        }
                        currentSection->insertEntry(currentEntry, NO_CHANGES);

                        break;
                    }
                }
            }

            ini->fixState();
            return ini;
        }

        // -------------------------------------------------------------------
        //      void writeInifile(const string& filename, InifilePtr ini)
        // -------------------------------------------------------------------
        void writeInifile(const string& filename, InifilePtr ini) {
            ofstream out(filename.c_str());
            if (!out.good()) throw IOError(filename);
            out << *ini;
            ini->fixState(); // "state in file == state in memory"
        }

        // --------------------------------------------------------------------------------------------------
        //      InifilePtr readOrCreateInifile(const string& inifilename, InifileCompleter complete_ini)
        // --------------------------------------------------------------------------------------------------
        InifilePtr readOrCreateInifile(const string& inifilename, InifileCompleter complete_ini) {
            InifilePtr ini;

            try {
                ini = readInifile(inifilename);
                ini->fixState(); // "state in file == state in memory"
            }
            catch (Error &err) { // if file not found = > start with empty ini
                cerr << err.get_print_message() << "\n";
                cerr << "Creating new inifile '" << inifilename << "'\n";
                ini = new Inifile("<new inifile>");
            }

            complete_ini(ini);

            ChangesState changes = ini->changesMade();
            if (changes != NO_CHANGES) {
                writeInifile(inifilename, ini);
                if (changes==MAJOR_CHANGES) {
                    throw Error(false, string("'")+inifilename+"' was created or extended.\n"
                                "It's really recommended that you edit it now.");
                }
            }

            return ini;
        }
    };
};



