/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000
// Ralf Westram
// (Coded@ReallySoft.de)
//
// Permission to use, copy, modify, distribute and sell this software
// and its documentation for any purpose is hereby granted without fee,
// provided that the above copyright notice appear in all copies and
// that both that copyright notice and this permission notice appear
// in supporting documentation.  Ralf Westram makes no
// representations about the suitability of this software for any
// purpose.  It is provided "as is" without express or implied warranty.
//
// This code is part of my library.
// You may find a more recent version at http://www.reallysoft.de/
//
/////////////////////////////////////////////////////////////////////////////

#ifndef INIFILE_H
#define INIFILE_H

#ifndef __MAP__
#include <map>
#endif

#ifndef ERR_H
#include "Err.h"
#endif
#ifndef TOOLS_H
#include "Tools.h"
#endif
#ifndef SMARTPTR_H
#include <smartptr.h>
#endif

namespace rs {

    /// Classes to read/write inifiles
    namespace ini {

        //  ----------------------
        //      class Comment
        //  ----------------------

        const char comment_char = ';';

        /** @brief How to handle changes
            @li NO_CHANGES=ignore
            @li MINOR_CHANGES=save
            @li MAJOR_CHANGES=save & throw Error
        */
        enum ChangesState {
            NO_CHANGES,
            MINOR_CHANGES,
            MAJOR_CHANGES
        };

        extern bool need_check_for_changes;

        /// Comment in Inifile
        class Comment {
        private:
            std::string text;
            Comment *next;

            void split_multiline();

        public:
            /// construct a line of comment
            Comment(const std::string& text_) : text(text_), next(0) { split_multiline(); }
            virtual ~Comment() { delete next; }

            /// append @p comment to self
            void append(Comment *comment);

            /// append @p text_ to self
            void append(const std::string& text_) { append(new Comment(text_)); }

            /// write comment to stream
            void dump(std::ostream& out) const;
        };

        //  --------------------------
        //      class Commentable
        //  --------------------------

        /// Section or Entry - may have a comment
        class Commentable {
        private:
            Comment *comment;

        public:
            Commentable() : comment(0) {}
            virtual ~Commentable() { delete comment; }

            void append(Comment *cmt) {
                if (comment) comment->append(cmt);
                else comment = cmt;
            }

            /** append/create a comment
                @param text one line of comment
             */
            void append(const std::string& text) { append(new Comment(text)); }

            /// write comment to stream
            void dump(std::ostream& out) const {
                if (comment) {
                    out << '\n';
                    comment->dump(out);
                }
            }

            /** @brief test if there is a comment
                @return true, if a comment exists
            */
            bool comment_exists() const { return comment != 0; }
        };

#define UNDEFINED_FILEPOSITION (-1UL)

        // ----------------------------
        //      class FilePosition
        // ----------------------------
        /// stores the line number where a Section or Entry has been defined
        class FilePosition {
        private:
            unsigned long lineNumber;

        public:
            /// create an undefined FilePosition
//             FilePosition() : lineNumber(UNDEFINED_FILEPOSITION) {}

            /** @brief create a defined FilePosition
                @param lineNumber_ the line number
            */
            FilePosition(unsigned long lineNumber_) : lineNumber(lineNumber_) {}
            virtual ~FilePosition() {}

            /// returns true if FilePosition is defined
            bool is_defined() const { return lineNumber != UNDEFINED_FILEPOSITION; }

            /// returns line number of FilePosition
            unsigned long LineNumber() const { assert(is_defined()); return lineNumber; }

            void throw_Error(const std::string& filename, const std::string& message) const;
        };


        // --------------------------------------------------------------------------------
        //     class Entry
        // --------------------------------------------------------------------------------

        class Section;

        /// stores one inifile entry
        class Entry : public Commentable, public FilePosition {
        private:
            std::string  name;
            std::string  content;
            bool         changed;
            Section     *my_section;

            /// mark entry as 'changed'
            void set_changed() { changed = true; }

            friend class Section;

        public:

            /** @brief create a new Entry
                @param content_ default content
                @param name_ name of the Entry
                @param line_number line number in the inifile
            */
            Entry(const std::string& name_, const std::string& content_, unsigned long line_number = UNDEFINED_FILEPOSITION)
                : FilePosition(line_number), name(name_), content(content_), changed(false), my_section(0)
            {}
            virtual ~Entry() {}

            /** @brief get pointer to my Section
                @return pointer to my Section (or NULL if Entry was not yet inserted into a Section)
            */
            const Section *mySection() const { return my_section; }

            /** @brief test if Entry was changed
                @return true if Entry was changed
            */
            bool changesMade() const { return changed; }

            /** @brief get name of entry
                @return the name of the Entry
            */
            const std::string& getName() const { return name; }

            /// get current value of the entry
            const std::string& getContent() const { return content; }

            /// set new value
            void setContent(const std::string& content_) {
                if (content_ != content) {
                    content = content_;
                    set_changed();
                }
            }

            /** @brief get current value of the entry as numeric value
                @exception Error if value is non-numeric
            */
            double getNumericContent() const;

            /** @brief get current value of the entry as numeric integer value
                @exception Error if value is non-numeric
            */
            long getIntegerContent() const;


            /** @brief sets the value of the Entry to a numeric value
                @param content_ the numeric value
            */
            void setNumericContent(double content_) {
                setContent(rs::str::strf("%f", content_));
            }

            /// comment out obsolete entry
            void mark_obsolete() { name = std::string("obsoleteEntry_")+name; set_changed(); }

            /// compare two Entry's
            bool operator<(const Entry& other) const { return str::stricmp(name, other.name) < 0; }

            /// write Entry to stream
            void dump(std::ostream& out) const {
                Commentable::dump(out);
                out << getName() << '=' << getContent() << "\n";
            }
//             virtual void dump(ostream& out) const = 0;

            /// @if 0
            DECLARE_STREAM_OP(Entry);
            /// @endif

            /// throws an error referring the entry
            void throw_Error(const std::string& message) const;
        };

        /// SmartPtr to Entry
        typedef SmartPtr<Entry> EntryPtr;

        // --------------------------------------------------------------------------------
        //     class Section
        // --------------------------------------------------------------------------------

        /// sort predicate used to sort map's case-insensitive
        struct ltistr {
            bool operator()(const std::string& s1, const std::string& s2) const {
                return s1<s2;
            }
        };

        /// Stores several Entry's
        typedef std::map<std::string, EntryPtr, ltistr > Entries;

        class Inifile;

        /// contains a whole section
        class Section : public Commentable, public FilePosition {
        private:
            std::string           name;
            Entries               entries;
            mutable ChangesState  changes_made;
            Inifile              *my_inifile;

            /** @brief Update chages-state of Section
                @param mode minimum wanted mode
                @note the current mode remains unchanged, if it is bigger than @p mode.
            */
            void set_changes_made(ChangesState mode) {
                if (mode>changes_made) {
                    changes_made           = mode;
                    need_check_for_changes = true;
                }
            }

            friend class Inifile;

            EntryPtr expectEntry(const std::string& name_) const;

        public:
            /// create an empty section named @p name_
            Section(const std::string& name_, unsigned long line_number) : FilePosition(line_number), name(name_), changes_made(NO_CHANGES), my_inifile(0) {}
            virtual ~Section() {}

            /** @brief get pointer to my Section
                @return pointer to my Section (or NULL if Entry was not yet inserted into a Section)
            */
            const Inifile *myInifile() const { return my_inifile; }

            /// get name of the Section
            const std::string& getName() const { return name; }

            /** @brief insert an entry into Section
                @param entry entry to insert
                @param mode How to handle changes (defaults to MINOR_CHANGES).
            */
            void insertEntry(EntryPtr entry, ChangesState mode=MINOR_CHANGES) {
                assert(entry->my_section == 0);
                entry->my_section     = this;
                entry->set_changed();

                entries[entry->getName()] = entry;
                set_changes_made(mode);
            }

            /** @brief search for an entry
                @param name_ name of entry
                @return entry (if entry was not found a Null SmartPtr is returned)
             */
            const EntryPtr findEntry(const std::string& name_) const {
                return const_cast<Section&>(*this).findEntry(name_);
            }
            /// @overload
            EntryPtr findEntry(const std::string& name_) {
                Entries::iterator e = entries.find(name_);
                if (e==entries.end()) return EntryPtr();
                return e->second;
            }

            /** @brief search for or create an Entry
                @param name_ name of Entry
                @param content if the Entry does not exist this is the default value
                @param mode How to handle changes (defaults to MINOR_CHANGES).
                @param add_comment Comment added if entry was created.
                @return entry
            */
            EntryPtr findOrCreateEntry(const std::string& name_, const std::string& content, ChangesState mode=MINOR_CHANGES, const char *add_comment=0);

            /// @overload
            EntryPtr findOrCreateEntry(const std::string& name_, long default_value, ChangesState mode=MINOR_CHANGES, const char *add_comment=0);

            /** @brief search for or create an Entry
                @param name_ name of Entry
                @param get_content if entry does not exist, the user function @p get_content
                will be called and the result will be used as default value
                @param mode How to handle changes (defaults to MINOR_CHANGES).
                @param add_comment Comment added if entry was created.
                @return entry
                @overload
            */
            EntryPtr findOrCreateEntry(const std::string& name_, std::string (*get_content)(), ChangesState mode = MINOR_CHANGES, const char *add_comment=0);

            /** @brief Searches section for an existing Entry
                @param name_ Name of the entry
                @return The value of the Entry.
                @exception Error if no such Entry was found.
            */
            const std::string& findStringEntry(const std::string& name_) const;

            /** @brief Searches section for an existing Entry
                @param name_ Name of the entry
                @return The numeric value of the Entry.
                @exception Error if no such Entry was found or if content is not numeric.
            */
            double findNumericEntry(const std::string& name_) const;

            /** @brief Searches section for an existing Entry
                @param name_ Name of the entry
                @return The numeric integer value of the Entry.
                @exception Error if no such Entry was found or if content is not numeric.
            */
            long findIntegerEntry(const std::string& name_) const;

            /// first section
            Entries::const_iterator begin() const { return entries.begin(); }
            /// behind last section
            Entries::const_iterator end() const { return entries.end(); }

            /// delete all entries from section
            void clearEntries() { entries.clear(); }

            /// write Section to stream
            void dump(std::ostream& out) const {
                Commentable::dump(out);
                out << '[' << name << ']' << "\n";
                for (Entries::const_iterator e=entries.begin(); e!=entries.end(); ++e) {
                    out << *(e->second);
                }
                out << "\n";
            }
            /// @if 0
            DECLARE_STREAM_OP(Section);
            /// @endif

            /** @brief test if changes were made in this section
                @return current ChangesState
            */
            ChangesState changesMade() const;

            /// reset ChangesState to NO_CHANGES
            void fixState() { changes_made = NO_CHANGES; }

            /// throws an error referring the section
            void throw_Error(const std::string& message) const;
        };

        /// SmartPtr to Section
        typedef SmartPtr<Section> SectionPtr;

        /// mapping for sections
        typedef std::map<std::string, SectionPtr, ltistr > Sections;

        // -----------------------
        //      class Inifile
        // -----------------------

        /// a whole inifile
        class Inifile
        {
        private:
            /// contains all Section's of the Inifile
            Sections sections;

            /// records if changes were made to the Inifile
            mutable ChangesState changes_made;

            std::string loaded_from_file;

            /** @brief update ChangesState of Inifile
                @param mode new ChangesState
                @note the current mode remains unchanged, if it is bigger than @p mode.
            */
            void set_changes_made(ChangesState mode) { if (mode>changes_made) changes_made = mode; }

        public:
            /** @brief create an empty inifile
                @see ini::readInifile
                @see ini::readOrCreateInifile
            */
            Inifile(const std::string& loaded_from_file_) : changes_made(NO_CHANGES), loaded_from_file(loaded_from_file_) {}
            virtual ~Inifile() {}

            /// start-iterator over all sections of the inifile
            Sections::const_iterator begin() const { return sections.begin(); }

            /// end-iterator over all sections of the inifile
            Sections::const_iterator end() const { return sections.end(); }

            /// returns the filename the inifile was loaded from
            const std::string& loadedFrom() const { return loaded_from_file; }

            /** @brief insert a new Section to Inifile
                @param section Section to insert.
                @param mode How to handle changes (defaults to MINOR_CHANGES).
            */
            void insertSection(SectionPtr section, ChangesState mode=MINOR_CHANGES) {
                assert(section->my_inifile == 0);
                section->my_inifile = this;
                section->set_changes_made(mode);

                sections[section->getName()] = section;
                set_changes_made(mode);
            }

            /** @brief search a Section
                @param name_ name of section
                @return section (if section was not found Null SectionPtr is returned)
            */
            const SectionPtr findSection(const std::string& name_) const {
                return const_cast<Inifile&>(*this).findSection(name_);
            }
            /// @overload
            SectionPtr findSection(const std::string& name_);

            /** @brief find an existing Section
                @param name_ name of section
                @return section (if section was not found an Error is thrown)
            */
            SectionPtr expectSection(const std::string& name_) const;

            /** @brief search for or create a Section

                if Section does not exist, it is created.

                @param name_ name of section
                @param mode How to handle changes (defaults to MINOR_CHANGES).
                @return found or created section
            */
            SectionPtr findOrCreateSection(const std::string& name_, ChangesState mode=MINOR_CHANGES);

            /** @brief search an entry

                @param section name of section
                @param entry name of entry
                @return entry
                @exception Error if section or entry was not found
                @see Section::findEntry
            */
            const EntryPtr findEntry(const std::string& section, const std::string& entry) const;

            /// @overload
            EntryPtr findEntry(const std::string& section, const std::string& entry) {
                return const_cast<const Inifile*>(this)->findEntry(section, entry);
            }

            /** @brief search a string entry
                @param section name of section
                @param entry name of entry
                @return the string contained in the entry
                @exception Error if section or entry does not exist or entry is no string entry
                @see findSection
                @see Section::findStringEntry
            */
            const std::string& findStringEntry(const std::string& section, const std::string& entry) const;

            /** @brief search a numeric entry
                @param section name of section
                @param entry name of entry
                @return the numeric value contained in the entry
                @exception Error if section or entry does not exist or entry is no numeric entry
                @see findSection
                @see Section::findNumericEntry
            */
            double findNumericEntry(const std::string& section, const std::string& entry) const;

            /** @brief search a numeric entry
                @param section name of section
                @param entry name of entry
                @return the numeric value contained in the entry
                @exception Error if section or entry does not exist or entry is no numeric entry
                @see findSection
                @see Section::findIntegerEntry
            */
            long findIntegerEntry(const std::string& section, const std::string& entry) const;

            /// write Inifile to stream
            void dump(std::ostream& out) const {
                for (Sections::const_iterator s = sections.begin(); s!=sections.end(); ++s) {
                    out << *(s->second);
                }
            }
            /// @if 0
            DECLARE_STREAM_OP(Inifile);
            /// @endif

            /** @brief test if changes were made
                @return info about made changes
                @see ChangesState
            */
            ChangesState changesMade() const;

            /** @brief set everything to unchanged
                @note Use this after saving the inifile to disk with @ref writeInifile
            */
            void fixState();
        };

        /// SmartPtr to Inifile
        typedef SmartPtr<Inifile> InifilePtr;

        /** @brief read inifile from %file
            @param filename name of inifile
            @return Inifile
            @exception Error
        */
        InifilePtr readInifile(const std::string& filename);

        /** @brief save inifile to %file
            @param filename name of inifile
            @param ini inifile to save
            @exception Error (in case of I/O problems)
            @note Use readOrCreateInifile to open an inifile
            @note Calls @ref Inifile::fixState
        */
        void writeInifile(const std::string& filename, InifilePtr ini);

        /// function type used for readOrCreateInifile
        typedef void (*InifileCompleter)(InifilePtr);

        /** @brief read inifile or create from scratch
            @param inifilename name of inifile
            @param complete_ini user function which inserts default values into the inifile.
            This is called in both cases (read and create) and should look for
            all needed sections/entries. If a section (or entry) does not exist it
            has to be created and inserted.
            @exception Error (if MAJOR_CHANGES were done in complete_ini() or in case of I/O problems)
            @see findOrCreateSection
            @see findOrCreateEntry
            @see readInifile
        */
        InifilePtr readOrCreateInifile(const std::string& inifilename, InifileCompleter complete_ini);

    };
};

#else
#error inifile.h included twice
#endif // INIFILE_H
