/////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) 2000
// Ralf Westram
// (Coded@ReallySoft.de)
//
// Permission to use, copy, modify, distribute and sell this software
// and its documentation for any purpose is hereby granted without fee,
// provided that the above copyright notice appear in all copies and
// that both that copyright notice and this permission notice appear
// in supporting documentation.  Ralf Westram makes no
// representations about the suitability of this software for any
// purpose.  It is provided "as is" without express or implied warranty.
//
// This code is part of my library.
// You may find a more recent version at http://www.reallysoft.de/
//
/////////////////////////////////////////////////////////////////////////////

#ifndef RS_ASSERT_H
#define RS_ASSERT_H

#ifndef __STDEXCEPT__
#include <stdexcept>
#endif
#ifndef __CASSERT__
#include <cassert>
#endif

#ifndef RSDEF_H
#include "rsdef.h"
#endif

namespace rs {

    /** @brief installs a terminator

        When you define a @b global instance of Terminator you get a warning if uncaught exceptions were thrown.

        @note @ref cmdline::ProgramInfo installs a Terminator when init() is called.
        @see handle_exception
    */
    class Terminator {
    private:

    public:
        Terminator();
        virtual ~Terminator();
    };

    /** @brief cerr printing exception handler

        @par Usage example:
        @code
        Terminator my_terminator;

        int main(void) {
        try {

        return EXIT_SUCCESS;
        }
        catch (exception& e) { handle_exception(e); }
        catch (...) { Error e("Unknown exception"); e.print(cerr); }

        return EXIT_FAILURE;
        }
        @endcode
    */
    void handle_exception(std::exception& e);
};

// remove assert from <cassert>
#undef assert

#if defined(DEBUG)

#define RAISE_EXCEPTION *(char*)0 = 0

namespace rs {
    /// debug function (used from assert and UNGUARDED)
    inline void guard() {
        std::cout << "guard\n";
    }
};

#define blub(cond)                                                                      \
do {                                                                                    \
    if (!(cond)) { std::cerr << rs::assertionText(#cond, __FILE__, __LINE__) << "\n";   \
    rs::guard();                                                                        \
    RAISE_EXCEPTION;                                                                    \
    exit(1);                                                                            \
    }                                                                                   \
} while (0)

#define assert(bed) blub(bed)

#define hassert(bed) assert(bed)

#define GUARDED try {
#define UNGUARDED }                                             \
catch (exception& e) {                                          \
    try {                                                       \
        reThrow(e);                                             \
    }                                                           \
    catch (Error& err) {                                        \
        cerr << "Guard sees exception: " << err.get_message()   \
              << " " << __FILE__ << '#' << __LINE__ << endl;    \
        assert(0);                                              \
    }                                                           \
}                                                               \
catch (...) {                                                   \
    cerr << "Guard sees unknown exception"                      \
         << " " << __FILE__ << '#' << __LINE__ << endl;         \
    assert(0);                                                  \
}

#else // !DEBUG

/// assertion (tested in DEBUG-Version)
#define assert(bed)

/** @brief assertion (tested in DEBUG and NDEBUG mode)

    @li In DEBUG mode @ref assert() is called (so you easily find the location with the debugger)
    @li In NDEBUG mode a logic_error is thrown
*/
#define hassert(bed) do { if (!(bed)) throw std::runtime_error(rs::assertionText(#bed, __FILE__, __LINE__)); } while(0)

#define GUARDED {
#define UNGUARDED }

#endif // DEBUG

/// nonnull(x) -- @ref hassert for expression x
template <class T> const T& nonnull(const T& t) {
    if (t == 0) hassert(t);
    return t;
}
template <class T> T& nonnull(T& t) {
    if (t == 0) hassert(t);
    return t;
}

#else
#error rs_assert.h included twice
#endif // RS_ASSERT_H

